# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
import time


class UndoStack:

    def __init__(self):

        self.current = {}
        self.history = []
        self.future = []

    def set_current(self, key, state):

        self.current[key] = state, time.time()

    def add(self, key, state):

        self.future = []
        if key in (u'__experiment__', u'__newitem__'):
            self.history.append((key, state))
            return
        timestamp = time.time()
        if key in self.current:
            _state, _timestamp = self.current[key]
            if _state == state:
                return
            if _timestamp >= timestamp-1:
                self.current[key] = state, timestamp
                return
            self.history.append((key, self.current[key][0]))
        self.current[key] = state, timestamp
        if len(self.history) > cfg.max_undo_history:
            self.history = self.history[-cfg.max_undo_history:]

    def can_undo(self):

        return bool(self.history)

    def can_redo(self):

        return bool(self.future)

    def __len__(self):

        return len(self.history)

    def pop(self, l1, l2):

        if not l1:
            return None, None
        key, state = l1.pop()
        # Both undoing experiment changes and adding new items clears the
        # future.That is, these cannot be redone.
        if key == u'__experiment__':
            self.future = []
            _key, _state = l1.pop()
            if _key != u'__experiment__':
                return None, None
            return _key, _state
        if key == u'__newitem__':
            self.future = []
            return u'__newitem__', state
        l2.append((key, self.current[key][0]))
        self.current[key] = state, time.time()
        return key, state

    def undo(self):

        return self.pop(self.history, self.future)

    def redo(self):

        return self.pop(self.future, self.history)

    def peek(self, i=-1):

        if not(self.history):
            return None, None
        try:
            return self.history[i]
        except IndexError:
            return None, None



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import re
import os.path
import platform
import sys
import site
from libopensesame import metadata
from libopensesame.oslogging import oslogger


def parse_environment_file():
    r"""Parses the environment.yaml file if it exists."""
    if not os.path.exists(u'environment.yaml'):
        return
    with safe_open(u'environment.yaml') as fd:
        d = safe_yaml_load(fd.read())
    # Convert all values from UTF8 to the filesystem encoding
    for key, val in d.items():
        if isinstance(val, str):
            d[key] = safe_str(safe_decode(val), enc=sys.getfilesystemencoding())
        elif isinstance(val, list):
            d[key] = [
                safe_str(safe_decode(i), enc=sys.getfilesystemencoding())
                for i in val
            ]
        else:
            raise TypeError('values in environment.yaml should be str or list')
    # The Python path is added to sys.path, the rest is added as an environment
    # variable.
    if u'PYTHON_PATH' in d:
        if isinstance(d[u'PYTHON_PATH'], str):
            sys.path = d[u'PYTHON_PATH'].split(';') + sys.path
        else:
            sys.path = d[u'PYTHON_PATH'] + sys.path
        del d[u'PYTHON_PATH']
    for key in (u'OPENSESAME_EXTENSION_PATH', u'OPENSESAME_PLUGIN_PATH'):
        if isinstance(d.get(key, None), list):
            d[key] = u';'.join(d[key])
    os.environ.update(d)


def change_working_dir():
    """A horrifyingly ugly hack to change the working directory under Windows"""
    import libqtopensesame.qtopensesame

    if os.name == "nt":
        try:
            # Extract the part of the description containing the path
            s = str(libqtopensesame.qtopensesame)
            i = s.find(u'from \'') + 6
            j = s.find(u'\'>\'') - 1
            s = s[i:j]
            # Go up the tree until the path of the current script
            while not os.path.exists(os.path.join(s, u'opensesame')) and \
                    not os.path.exists(os.path.join(s, u'opensesame.exe')) and \
                    not os.path.exists(os.path.join(s, u'opensesamerun')) and \
                    not os.path.exists(os.path.join(s, u'opensesamerun.exe')):
                s = os.path.dirname(s)
            os.chdir(s)
            if s not in sys.path:
                sys.path.append(s)
        except Exception as e:
            oslogger.debug(u'failed to change working directory: %s' % e)


def opensesamerun_options():
    """Parse the command line options for opensesamerun"""
    import optparse

    parser = optparse.OptionParser(
        u'usage: opensesamerun [experiment] [options]', version=u'%s \'%s\'' %
        (metadata.__version__, metadata.codename))
    parser.set_defaults(subject=0)
    parser.set_defaults(logfile=None)
    parser.set_defaults(debug=False)
    parser.set_defaults(fullscreen=False)
    parser.set_defaults(width=1024)
    parser.set_defaults(height=768)
    parser.set_defaults(custom_resolution=False)
    group = optparse.OptionGroup(parser, u'Subject and log file options')
    group.add_option(u"-s", u"--subject", action=u"store", dest=u"subject",
                     help=u"Subject number")
    group.add_option(u"-l", u"--logfile", action=u"store", dest=u"logfile",
                     help=u"Logfile")
    parser.add_option_group(group)
    group = optparse.OptionGroup(parser, u"Display options")
    group.add_option(u"-f", u"--fullscreen", action=u"store_true",
                     dest="fullscreen", help=u"Run fullscreen")
    group.add_option(u"-c", u"--custom_resolution", action=u"store_true",
                     dest=u"custom_resolution", help=u"Do not use the display resolution specified in the experiment file")
    group.add_option(u"-w", u"--width", action=u"store",
                     dest=u"width", help=u"Display width")
    group.add_option(u"-e", u"--height", action=u"store", dest=u"height",
                     help=u"Display height")
    parser.add_option_group(group)
    group = optparse.OptionGroup(parser, u"Miscellaneous options")
    group.add_option(u"-d", u"--debug", action=u"store_true", dest=u"debug",
                     help=u"Print lots of debugging messages to the standard output")
    group.add_option(u"--stack", action=u"store_true",
                     dest=u"stack", help=u"Print stack information")
    parser.add_option_group(group)
    parser.add_option_group(group)
    options, args = parser.parse_args(sys.argv)

    # Set the default logfile based on the subject nr
    if options.logfile is None:
        options.logfile = u"subject-%s.csv" % options.subject

    if len(sys.argv) > 1 and os.path.exists:
        options.experiment = sys.argv[1]
    else:
        options.experiment = u""

    try:
        options.subject = int(options.subject)
    except:
        parser.error(u"Subject (-s / --subject) should be numeric")

    try:
        options.width = int(options.width)
    except:
        parser.error(u"Width (-w / --width) should be numeric")

    try:
        options.height = int(options.height)
    except:
        parser.error(u"Height (-e / --height) should be numeric")

    return options


def opensesamerun_ready(options):
    """
    Check if the opensesamerun options are sufficiently complete to run the
    experiment

    Arguments:
    options -- a dictionary containing the options

    Returns:
    True or False, depending on whether the options are sufficient
    """
    # Check if the experiment exists
    if not os.path.exists(options.experiment):
        return False

    # Check if the logfile is writable
    try:
        open(options.logfile, u"w")
    except:
        return False

    # Ready to run!
    return True


def messagebox(title, msg):
    """
    Presents a simple tk messagebox

    Arguments:
    title -- the title of the messagebox
    msg -- the message
    """
    import tkinter as Tkinter
    root = Tkinter.Tk()
    root.title(title)
    l = Tkinter.Label(root, text=msg, justify=Tkinter.LEFT, padx=8, pady=8,
                      wraplength=300)
    l.pack()
    b = Tkinter.Button(root, text=u"Ok", command=root.quit)
    b.pack(side=Tkinter.RIGHT)
    root.mainloop()


def strip_tags(s):
    """
    Strip html tags from a string and convert breaks to newlines.

    Arguments:
    s -- the string to be stripped

    Returns:
    The stripped string
    """
    import re
    return re.compile(r'<.*?>').sub(
        '', str(s).replace("<br />","\n").replace("<br>", "\n"))


def home_folder():
    """
    Determines the home folder.

    Returns:
    A path to the home folder.
    """
    import platform
    if platform.system() == u"Windows":
        home_folder = os.environ[u"APPDATA"]
    elif platform.system() == u"Darwin":
        home_folder = os.environ[u"HOME"]
    elif platform.system() == u"Linux":
        home_folder = os.environ[u"HOME"]
    else:
        home_folder = os.environ[u"HOME"]
    return safe_decode(home_folder, enc=sys.getfilesystemencoding())


def opensesame_folder():
    """
    Determines the folder that contains the OpenSesame executable. This is only
    applicable under Windows.

    Returns:
    The OpenSesame folder or None if the os is not Windows.
    """
    # Determines the directory name of the script or the directory name
    # of the executable
    #
    # There are two scenarios:
    #
    # - OpenSesame is run from source, in which case we go to the OpenSesame
    #   folder by going a levels up from the __file__ folder.
    # - OpenSesame is run in Anaconda, in which case we need to go two levels
    #   up to get out of the site-packages folder.
    if platform.system() == 'Darwin':
        return os.getcwd()
    if platform.system() == 'Windows':
        # To get the opensesame folder, simply jump to levels up
        path = safe_decode(
            os.path.dirname(__file__),
            enc=sys.getfilesystemencoding()
        )
        # The current should not be the site-packages folder, which
        # happens on Anaconda if launched in multiprocess mode.
        path = os.path.normpath(os.path.join(path, '..'))
        if path.endswith('Lib\\site-packages'):
            path = os.path.normpath(os.path.join(path, '..', '..'))
        return path
    return None


def open_url(url):
    """
    Open a URL in an OS specific way. The URL can be a file, website, etc.

    Arguments:
    url -- a url
    """
    oslogger.info(u'opening %s' % url)
    if url.startswith(u'http://') or url.startswith(u'https://'):
        import webbrowser
        webbrowser.open(url)
        return
    import platform
    import subprocess
    if platform.system() == u"Linux":
        pid = subprocess.Popen([u"xdg-open", url]).pid
    elif platform.system() == u"Darwin":
        pid = subprocess.Popen(["open", url]).pid
    elif platform.system() == u"Windows":
        try:
            os.startfile(url)
        except:
            oslogger.error(u"Failed to open '%s'" % url)
    else:
        oslogger.error(u"Failed to open '%s'" % url)


def strip_html(s):
    """
    Strips basic HTML tags from a string.

    Arguments:
    s		--	A string to strip.

    Returns:
    A stripped string.
    """
    s = s.replace(u'<br />', u'\n')
    for tag in [u'<i>', u'</i>', u'<b>', u'</b>']:
        s = s.replace(tag, u'')
    return s


def escape_html(s):
    """
    Escapes a string so that it can be displayed as HTML. This is useful for
    example for tracebacks, which use <> characters.

    Arguments:
    s	--	A string to escape. We assume Unicode input. str objects may cause
                    decoding errors.

    Returns:
    An escaped string.
    """
    # Note that we need to replace the '&' first, otherwise we'll start escaping
    # the escaped characters.
    l = [(u'&', u'&amp;'), (u' ', u'&nbsp;'),
         (u'\t', u'&nbsp;&nbsp;&nbsp;&nbsp;'), (u'<', u'&lt;'),
         (u'>', u'&gt;')]
    for orig, new in l:
        s = s.replace(orig, new)
    return s


snake_case_pattern = re.compile(r'(?<!^)(?=[A-Z])')


def snake_case(s):
    """Turns CamelCase to snake_case"""
    return snake_case_pattern.sub('_', s).lower()
    
    
def camel_case(s):
    """Turns snake_case to CamelCase"""
    return ''.join(x.capitalize() or '_' for x in s.split('_'))


# Build a list of base folders, that is, folders that can have
# opensesame_resources, opensesame_plugins, or opensesame_extensions as
# subfolders. Which folders are available depends on the environment, but the
# following folders are scanned:
#
# - The working directory (cwd)
# - The folder that contains libopensesame (parent_folder)
# - The share folder within an Anaconda/ Miniconda environment
# - The userbase folder:
#		- ~/.local (on Ubuntu 16.04)
# - The user site-packages folder:
#		- ~/.local/lib/python[X]/site-packages (on Ubuntu 16.04)
# - The global site-packages folders (on Ubuntu 16.04):
#		- /usr/local/lib/python3.5/dist-packages
#		- /usr/lib/python3/dist-packages
#		- /usr/lib/python3.5/dist-packages
# - /usr/local/share
# - /usr/share

cwd = os.getcwd()
parent_folder = safe_decode(
    os.path.dirname(os.path.dirname(__file__)),
    enc=sys.getfilesystemencoding()
)
base_folders = [cwd, parent_folder]
# Locate Anaconda/Miniconda share
base_folders.append(os.path.join(
    safe_decode(
        os.path.dirname(os.path.dirname(sys.executable)),
        enc=sys.getfilesystemencoding()
    ),
    u"share"
))
if hasattr(site, u'getuserbase'):
    base_folders.append(os.path.join(
        safe_decode(site.getuserbase(), enc=sys.getfilesystemencoding()),
        u'share'
    ))
if hasattr(site, u'getusersitepackages'):
    base_folders.append(os.path.join(
        safe_decode(site.getusersitepackages(), enc=sys.getfilesystemencoding()),
        u'share'
    ))
if hasattr(site, u'getsitepackages'):
    base_folders += [
        os.path.join(
            safe_decode(folder, enc=sys.getfilesystemencoding()),
            u'share'
        )
        for folder in site.getsitepackages()
    ]
base_folders += [u'/usr/local/share', u'/usr/share']
base_folders = list(filter(os.path.exists, base_folders))
resource_folders = [
    os.path.join(path, u'opensesame_resources')
    for path in base_folders
]
if 'OPENSESAME_RESOURCES_PATH' in os.environ:
    resource_folders.append(os.environ['OPENSESAME_RESOURCES_PATH'])



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension, suspend_events
from libqtopensesame.items.qtstructure_item import qtstructure_item
from qtpy.QtWidgets import QMenu, QToolBar
from .undo_stack import UndoStack
import difflib
from libqtopensesame.misc.translate import translation_context
_ = translation_context('undo_manager', category='extension')


class UndoManager(BaseExtension):
    """An extension that implements undo/ redo."""
    
    def event_startup(self):
        """Initializes the extension on startup."""
        self.undo_action = self.qaction('edit-undo', _('Undo'), self.undo,
                                        tooltip=_('Undo most recent action'),
                                        shortcut='Alt+Ctrl+Z')
        # This extension has two actions instead of one. This is a hacky
        # solution to insert the second action after the first one in both
        # the toolbar and menu.
        menu_pos = self.menu_pos()
        if menu_pos is not None:
            submenu_text, index, separator_before, separator_after = menu_pos
            submenu = self.get_submenu(submenu_text)
            submenu.insertAction(self.action, self.undo_action)
        if self.toolbar_pos() is not None:
            self.toolbar.insertAction(self.action, self.undo_action)
        self.initialize()

    def event_open_experiment(self, path):
        """Re-initializes when a new experiment is opened."""
        self.initialize()

    def event_change_item(self, name):
        """Remember an items state when it has changed."""
        self.remember_item_state(name)

    def event_new_item(self, name, _type):
        """Remember addition of a new item."""
        self.remember_new_item(name)

    def activate(self):
        """Activated the redo action (undo is added as custom action in
        event_startup).
        """
        self.redo()

    def initialize(self):
        """Initializes an empty undo stack."""
        self.stack = UndoStack()
        for item in self.experiment.items:
            self.remember_item_state(item)
        self.undo_action.setEnabled(self.stack.can_undo())

    def remember_experiment_state_before(self):
        """Remember the entire experiment state before it has changed."""
        self.remember_experiment_state()

    def remember_experiment_state_after(self):
        """Remember the entire experiment state after it has changed, and
        update the current-item status. If no change has actually occurred the
        last two experiments are discarded.
        """
        self.remember_experiment_state()
        if self.prune_unchanged_experiment():
            return
        for item in self.experiment.items:
            self.make_current(item)

    def make_current(self, item):
        """Update the current-item status for one item.

        Parameters
        ----------
        item
            The item name.
        """
        script = self.experiment.items[item].cached_script
        self.stack.set_current(item, script)

    def remember_item_state(self, item):
        """Push the state of an item to the undo stack.

        Parameters
        ----------
        item
            The item name.
        """
        script = self.experiment.items[item].cached_script
        self.stack.add(item, script)
        self.set_enabled(self.stack.can_redo())
        self.undo_action.setEnabled(self.stack.can_undo())

    def remember_experiment_state(self):
        """Push the state of the entire experiment to the undo stack."""
        script = self.experiment.to_string()
        self.stack.add('__experiment__', script)
        self.set_enabled(self.stack.can_redo())
        self.undo_action.setEnabled(self.stack.can_undo())

    def remember_new_item(self, name):

        self.stack.add('__newitem__', name)
        self.make_current(name)
        self.set_enabled(self.stack.can_redo())
        self.undo_action.setEnabled(self.stack.can_undo())

    def prune_unchanged_experiment(self):
        """If the last two items on the undo stack are full experiments, and
        there is no change between them, discard them.
        """
        if len(self.stack) < 2:
            return False
        key1, state1 = self.stack.peek(-1)
        key2, state2 = self.stack.peek(-2)
        if key1 == key2 == '__experiment__' and state1 == state2:
            self.stack.history = self.stack.history[:-2]
            return True
        return False

    @suspend_events
    def undo(self, dummy=False):
        """Perform an undo action."""
        if not self.stack:
            return
        item, script = self.stack.undo()
        if item == '__experiment__':
            self.main_window.regenerate(script)
        elif item == '__newitem__':
            del self.experiment.items[script]
            self.tabwidget.close_all()
        else:
            if item not in self.experiment.items:
                return
            self.experiment.items[item].from_string(script)
            self.experiment.items[item].update()
            self.experiment.items[item].open_tab()
            if isinstance(self.experiment.items[item], qtstructure_item):
                self.experiment.items.clear_cache()
                self.experiment.build_item_tree()
        self.set_enabled(self.stack.can_redo())
        self.undo_action.setEnabled(self.stack.can_undo())

    @suspend_events
    def redo(self):
        """Perform a redo action."""
        item, script = self.stack.redo()
        if item not in self.experiment.items:
            return
        self.experiment.items[item].from_string(script)
        self.experiment.items[item].update()
        self.experiment.items[item].open_tab()
        if isinstance(self.experiment.items[item], qtstructure_item):
            self.experiment.items.clear_cache()
            self.experiment.build_item_tree()
        self.set_enabled(self.stack.can_redo())
        self.undo_action.setEnabled(self.stack.can_undo())

    def show(self):
        """Print a summary of undo actions to the debug window."""
        item, old_script = self.stack.peek()
        if item is None:
            print('\nThe undo stack is empty')
            return
        self.console.write('\nMost recent: %s\n\n' % item)
        if item == '__experiment__':
            old_script = self.stack.peek(-2)[1]
            new_script = self.stack.peek(-1)[1]
        elif item == '__newitem__':
            new_script = old_script
        else:
            new_script = self.stack.current[item][0]
        for line in difflib.ndiff(old_script.splitlines(),
                                  new_script.splitlines()):
            if line.startswith('+'):
                self.console.write('\x1b[32;1m')
            elif line.startswith('-'):
                self.console.write('\x1b[31;1m')
            else:
                continue
            self.console.write(line + '\n')
        self.console.write('\x1b[0m')
        self.console.write('\nFull stack:\n\n')
        for i, (item, script) in enumerate(self.stack.history[::-1]):
            self.console.write('%d - %s\n' % (i, item))

    # All these events simply undo by restoring the complete experiment state.
    # This is crude, but works for now.

    def event_prepare_regenerate(self):
        self.extension_manager.suspend_until('regenerate')
        self.remember_experiment_state_before()

    def event_regenerate(self):
        self.remember_experiment_state_after()

    def event_prepare_delete_item(self, name):
        self.extension_manager.suspend_until('delete_item')
        self.remember_experiment_state_before()

    def event_delete_item(self, name):
        self.remember_experiment_state_after()
        
    def event_prepare_move_item(self, name):
        self.extension_manager.suspend_until('move_item')
        self.remember_experiment_state_before()

    def event_move_item(self, name):
        self.remember_experiment_state_after()
        
    def event_prepare_linked_copy(self, name):
        self.extension_manager.suspend_until('linked_copy')
        self.remember_experiment_state_before()

    def event_linked_copy(self, name):
        self.remember_experiment_state_after()

    def event_prepare_rename_item(self, from_name, to_name):
        self.extension_manager.suspend_until('rename_item')
        self.remember_experiment_state_before()

    def event_rename_item(self, from_name, to_name):
        self.remember_experiment_state_after()

    def event_prepare_purge_unused_items(self):
        self.extension_manager.suspend_until('purge_unused_items')
        self.remember_experiment_state_before()

    def event_purge_unused_items(self):
        self.remember_experiment_state_after()

    def event_prepare_change_experiment(self):
        self.extension_manager.suspend_until('change_experiment')
        self.remember_experiment_state_before()

    def event_change_experiment(self):
        self.remember_experiment_state_after()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.mouse_response import MouseResponse as MouseResponseRuntime
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.validators import TimeoutValidator
from libqtopensesame.misc.translate import translation_context
from libqtopensesame._input.item_combobox import ItemComboBox
from libqtopensesame.items.feedpad import Feedpad
_ = translation_context(u'mouse_response', category=u'item')


class MouseResponse(MouseResponseRuntime, QtPlugin):

    """mouse_response item GUI"""
    description = _(u'Collects mouse responses')
    help_url = u'manual/response/mouse'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        r"""Constructor

        Parameters
        ----------
        name
            The item name.
        experiment
            The experiment object.
        string, optional
            The item definition string, or None for default initialization.
        """
        MouseResponseRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def init_edit_widget(self):
        r"""Initialize controls."""
        super().init_edit_widget(stretch=True)
        self.add_line_edit_control(
            var=u'correct_response',
            label=_(u'Correct response'),
            info=_(u'Leave empty to use "correct_response"')
        )
        self.add_line_edit_control(
            var=u'allowed_responses',
            label=_(u'Allowed responses'),
            info=_(u'Separated by semicolons, e.g. "left_button;right_button"')
        )
        self.add_line_edit_control(
            var=u'timeout',
            label=_(u'Timeout'),
            tooltip=_(u'In milliseconds or "infinite"'),
            validator=TimeoutValidator(self)
        )
        self._combobox_sketchpad = ItemComboBox(
            self.main_window,
            filter_fnc=lambda item: isinstance(
                self.experiment.items[item],
                Feedpad
            )
        )
        self._combobox_sketchpad.activated.connect(self.apply_edit_changes)
        self.auto_combobox[u'linked_sketchpad'] = self._combobox_sketchpad
        self.add_control(
            label=_(u'Linked sketchpad'),
            widget=self._combobox_sketchpad,
            info=_('Elements define regions of interest')
        )
        self.add_combobox_control(
            var=u'event_type',
            label=_(u'Event type'),
            options=[
                u'mouseclick',
                u'mouserelease'
            ]
        )
        self.add_checkbox_control(
            var=u'show_cursor',
            label=_(u'Visible mouse cursor')
        )
        self.add_checkbox_control(
            var=u'flush',
            label=_(u'Flush pending mouse clicks')
        )

    def edit_widget(self):
        self._combobox_sketchpad.refresh()
        QtPlugin.edit_widget(self)

    def rename(self, from_name, to_name):
        QtPlugin.rename(self, from_name, to_name)
        if self.var.linked_sketchpad == from_name:
            self.var.linked_sketchpad = to_name

    def delete(self, item_name, item_parent=None, index=None):
        if self.var.linked_sketchpad == item_name:
            self.var.linked_sketchpad = u''


# Alias for backwards compatibility
mouse_response = MouseResponse



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
import traceback
import opensesame_extensions
from libopensesame.plugin_manager import PluginManager
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc.config import cfg
_ = translation_context(u'extension_manager', category=u'core')


def suspend_events(fnc):
    r"""A decorator that causes all the extension manager to be suspended while
    the function is executed.
    """
    def inner(self, *args, **kwdict):

        self.extension_manager.suspend()
        retval = fnc(self, *args, **kwdict)
        self.extension_manager.resume()
        return retval

    return inner


class ExtensionManager(BaseSubcomponent):

    r"""Initializes GUI extensions, and distributes signals (fires) to the
    relevant extensions.
    """
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main-window object.
        """
        self.setup(main_window)
        self.main_window.set_busy()
        QtWidgets.QApplication.processEvents()
        self._extensions = []
        self.events = {}
        self.provides = {}
        self._suspended = False
        self._suspended_until = None
        def extension_filter(ext_name): return False
        for ulext in self.unloaded_extension_manager.filter(
                modes=self.main_window.mode):
            # Ignoring disabled extensions
            cfg_key = f'plugin_enabled_{ulext.name}'
            if cfg_key in cfg and not cfg[cfg_key]:
                continue
            if extension_filter(ulext.name):
                oslogger.debug(u'filtering extension {}'.format(ulext.name))
                continue
            try:
                ext = ulext.build(self.main_window)
            except Exception as e:
                oslogger.error(f'Failed to load extension {ulext.name}: {e}')
                traceback.print_exc()
                self.console.write(e)
            else:
                self._extensions.append(ext)
                self.register_extension(ext)
                if ext.extension_filter is not None:
                    extension_filter = ext.extension_filter
        self.main_window.set_busy(False)

    def register_extension(self, ext):

        oslogger.debug(f'installing extension {ext.name()}')
        for event in ext.supported_events():
            if event not in self.events:
                self.events[event] = []
            self.events[event].append(ext)
        for provide in ext.supported_provides():
            if provide in self.provides:
                oslogger.warning(
                    f'multiple extensions provide {provide}')
                continue
            self.provides[provide] = ext

    def __getitem__(self, extension_name):
        r"""Emulates a dict interface for retrieving extensions.

        Parameters
        ----------
        extension_name : str
            The extension name.

        Returns
        -------
        base_extension
        """
        for ext in self._extensions:
            if extension_name in ext.aliases:
                return ext
        raise KeyError(f'Extension {extension_name} does not exist')

    def __contains__(self, extension_name):
        """
        Returns
        -------
        bool
            True if the extension exists, False otherwise.
        """
        return any(extension_name in ext.aliases for ext in self._extensions)

    def fire(self, event, **kwdict):
        r"""Fires an event to all extensions that support the event.

        Parameters
        ----------
        event
            The event name.
        **kwdict : dict
            A dictionary with keywords that are applicable to a particular
            event.
            
        Returns
        -------
        dict
            A dict with extension names as keys and returns values as values.
            This dict contains only return values that were not None.
        """
        if self._suspended_until == event:
            self._suspended = False
            self._suspended_until = None
        if self._suspended:
            oslogger.debug('event suspended {}'.format(event))
            return
        oslogger.debug('firing {}'.format(event))
        if event == u'open_experiment':
            for ext in self._extensions:
                ext.register_ui_files()
        return_values = {}
        for ext in self.events.get(event, []):
            try:
                return_value = ext.fire(event, **kwdict)
            except Exception as e:
                self.notify(
                    f'Extension {ext.name()} misbehaved on event {event} '
                    f'(see debug window for stack trace)',
                    category='warning')
                self.console.write(e)
                traceback.print_exc()
                oslogger.error(f'Extension {ext.name()} misbehaved on event '
                               f'{event}: {e}')
            else:
                if return_value is not None:
                    return_values[ext.name()] = return_value
        return return_values

    def provide(self, provide, **kwdict):

        if provide not in self.provides:
            oslogger.debug('cannot not provide {}'.format(provide))
            return None
        ext = self.provides[provide]
        oslogger.debug('{} provides {}'.format(ext, provide))
        try:
            return ext.provide(provide, **kwdict)
        except Exception as e:
            self.notify(
                f'Extension {self.name} misbehaved on providing {provide} '
                f'(see debug window for stack trace)',
                category='warning')
            self.console.write(e)
            traceback.print_exc()
            oslogger.error(f'Extension {ulext.name} misbehaved on providing '
                           f'{provide}: {e}')

    def activate(self, ext_name):

        try:
            ext = self[ext_name]
        except Exception:
            return
        ext.activate()

    def suspend(self):
        r"""Suspends all events."""
        self._suspended = True

    def suspend_until(self, event):
        r"""Suspends all events until a specific event is fired. This is useful
        for situations where you want to suppress all events between a starting
        and ending event.
        """
        self._suspended = True
        self._suspended_until = event

    def resume(self):
        r"""Resumes all events."""
        self._suspended = False



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import warnings
from numbers import Number
try:
    from types import NoneType
except ImportError:  # Python <= 3.9
    NoneType = type(None)
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidOpenSesameScript, OSException, \
    VariableDoesNotExist, InvalidValue


class VarStore:
    """__New in 4.0.0__: As of OpenSesame 4.0, all experimental variables are
    also available in the Python workspace. This means that you therefore 
    don't need the `var` object anymore.
    
    
    The `var` object provides access to experimental variables.
    Experimental variables are the variables that live in the GUI, and are
    commonly set as independent variables in the LOOP item, referred
    to using
    the square-bracket (`[my_variable]`) notation, and logged by
    the LOGGER
    item.

    A `var` object is created automatically when the experiment starts.
    In addition to the functions listed below, the following semantics are
    supported:

    __Example__:

    ~~~ .python
    # Set an experimental variable
    var.my_variable = u'my_value'
    # Get an experimental variable
    print(u'Subject nr = %d' % var.subject_nr)
    # Delete (unset) an experimental
    variable
    del var.my_variable
    # Check if an experimental variable exists
    if
    u'my_variable' in var:
        print(u'my_variable exists!')
    # Loop through all
    experimental variables
    for var_name in var:
            print(u'variable found:
    %s' % var_name)
    ~~~

    [TOC]
    """
    def __init__(self, item, parent=None):
        r"""Constructor.

        Parameters
        ----------
        item : item
            The associated item.
        parent : var_store, NoneType, optional
            The parent var_store (i.e. for the experiment) or `None` for no
            parent.
        """
        object.__setattr__(self, u'__item__', item)
        object.__setattr__(self, u'__parent__', parent)
        object.__setattr__(self, u'__vars__', {})
        object.__setattr__(self, u'__lock__', None)
        self._copy_class_description()

    def _copy_class_description(self):
        r"""Item descriptions can be class properties. In that case, copy theme
        into the var store to avoid warnings.
        """
        if hasattr(self.__item__.__class__, u'description'):
            self.__vars__[u'description'] = self.__item__.__class__.description

    def _check_var_name(self, var):
        r"""Checks whether a variable name is valid, and raises an
        Exception if it isn't.

        Parameters
        ----------
        var
            The variable name to check.
        """
        try:
            self.__item__.experiment
        except:
            return
        if isinstance(var, str) and \
                self.__item__.experiment.syntax.valid_var_name(var):
            return
        raise InvalidOpenSesameScript(f'"{var}" is not a valid variable name')

    def __contains__(self, var):
        r"""Implements the `in` operator to check if a variable exists."""
        self._check_var_name(var)
        if var in self.__vars__:
            return True
        if hasattr(self.__item__, var):
            warnings.warn(u'var %s is stored as attribute of item %s'
                          % (var, self.__item__.name))
            return True
        if self.__parent__ is not None:
            return self.__parent__.__contains__(var)
        return False

    def __delattr__(self, var):
        r"""Implements the `del` statement to delete a variable."""
        if var in self.__vars__:
            del self.__vars__[var]
        if hasattr(self.__item__, var):
            warnings.warn(u'var %s is stored as attribute of item %s'
                          % (var, self.__item__.name))
            delattr(self.__item__, var)

    def __getattr__(self, var):
        r"""Implements property retrieval to allow direct access to variables."""
        return self.get(var)

    def __setattr__(self, var, val):
        r"""Implements property assignment."""
        self.__vars__[var] = val

    def clear(self, preserve=[]):
        r"""*New in 3.1.2*

        Clears all experimentals variables.

        Parameters
        ----------
        preserve : list, optional
            A list of variable names that shouldn't be cleared.

        Examples
        --------
        >>> var.clear()
        """
        for var in list(self.__vars__.keys()):
            if var not in preserve:
                del self.__vars__[var]
        self._copy_class_description()

    def get(self, var, default=None, _eval=True, valid=None):
        r"""Gets an experimental variable.

        Parameters
        ----------
        var : str, unicode
            The variable to retrieve.
        default : any, optional
            A default value in case the variable doesn't exist, or `None` for
            no default value.
        _eval : bool, optional
            Determines whether the returned should be evaluated for variable
            references.
        valid : NoneType, list, optional
            A list of valid values, or `None` to allow all values.

        Examples
        --------
        >>> print('my_variable = %s' % var.get(u'my_variable'))
        >>> # Equivalent to:
        >>> print('my_variable = %s' % var.my_variable)
        >>> # But if you want to pass keyword arguments you need to use `get()`:
        >>> var.get(u'my_variable', default=u'a_default_value')
        """
        self._check_var_name(var)
        if self.__lock__ == var:
            raise OSException(f"Recursion detected! Is variable {var} defined "
                              f"in terms of itself (e.g., 'var = [var]')")
        if var in self.__vars__:
            val = self.__vars__[var]
        elif hasattr(self.__item__, var):
            warnings.warn(
                u'var %s is stored as attribute of item %s'
                % (var, self.__item__.name)
            )
            val = getattr(self.__item__, var)
        elif self.__parent__ is not None:
            val = self.__parent__.get(
                var,
                default=default,
                _eval=_eval,
                valid=valid
            )
        elif default is not None:
            val = default
        else:
            raise VariableDoesNotExist(var)
        if valid is not None and val not in valid:
            raise InvalidValue(
                f'Variable {var} should be in {valid}, not {val}')
        if _eval:
            object.__setattr__(self, u'__lock__', var)
            val = self.__item__.syntax.auto_type(
                self.__item__.syntax.eval_text(val))
            object.__setattr__(self, u'__lock__', None)
        return val

    def has(self, var):
        r"""Checks if an experimental variable exists.

        Parameters
        ----------
        var : str, unicode
            The variable to check.

        Examples
        --------
        >>> if var.has(u'my_variable'):
        >>>         print(u'my_variable has been defined!')
        >>> # Equivalent to:
        >>> if u'my_variable' in var:
        >>>         print(u'my_variable has been defined!')
        """
        return self.__contains__(var)

    def set(self, var, val):
        r"""Sets and experimental variable.

        Parameters
        ----------
        var : str, unicode
            The variable to assign.
        val : any
            The value to assign.

        Examples
        --------
        >>> var.set(u'my_variable', u'my_value')
        >>> # Equivalent to
        >>> var.my_variable = u'my_value'
        """
        self._check_var_name(var)
        if not isinstance(val, bool):
            # The logic here is that we first try to convert to float, and if this
            # is not successful, return the original type. If this is successful we
            # turn the float into an int if this doesn't result in data loss, and
            # otherwise return the float.
            try:
                val = float(val)
            except (TypeError, ValueError):
                pass
            try:
                ival = int(val)
            except (TypeError, ValueError, ArithmeticError):
                # A float can always be converted to an int, except nan values,
                # which result in ValueError, or inf values, which result in an
                # OverFlowError (which is a subclass of ArithmeticError).
                pass
            else:
                if ival == val:
                    val = ival
        self.__setattr__(var, val)

    def unset(self, var):
        r"""Deletes a variable.

        Parameters
        ----------
        var : str, unicode
            The variable to delete.

        Examples
        --------
        >>> var.unset(u'my_variable')
        >>> # Equivalent to:
        >>> del var.my_variable
        """
        self._check_var_name(var)
        self.__delattr__(var)

    def __iter__(self):
        r"""Implements the iterator."""
        return var_store_iterator(self)

    def __len__(self):
        r"""Returns the number of experimental variables that are stored in the
        `var_store` object.

        Returns
        -------
        int
            The number of experimental variables.
        """
        return len(self.__vars__)

    def vars(self):
        r"""Returns a list of experimental variables. Because experimental
        variables can be stored in multiple places, this list may not be
        exhaustive. That is, `u'my_var' in var` may return `True`, while
        u'my_var' is not in the list of variables as returned by this function.

        Returns
        -------
        list
            A list of variable names.

        Examples
        --------
        >>> for varname in var.vars():
        >>>         print(varname)
        """
        return sorted(list(self.__vars__.keys()))

    def items(self):
        r"""Returns a list of (variable_name, value) tuples. See `var.vars()`
        for a note about the non-exhaustiveness of this function.

        Returns
        -------
        list
            A list of (variable_name, value) tuples.

        Examples
        --------
        >>> for varname, value in var.items():
        >>>         print(varname, value)
        """
        return list(self.__vars__.items())
        
    def is_default_loggable(self, var, val):
        """Checks whether a value is loggable by default. This includes int,
        str, byes, float, bool, and None values, as well as any type that is 
        derived from numbers.Number. Variable names that start with '_' are
        not logged by default. {nodoc}

        Parameters
        ----------
        var : str
        val

        Returns
        -------
        bool
        """
        return not var.startswith('_') and \
            isinstance(val, (str, bytes, NoneType, Number))

    def inspect(self):
        r"""Generates a description of all experimental variables, both alive
        and hypothetical.

        Returns
        -------
        dict
            A dict where variable names are keys, and values are dicts with
            source, value, and alive keys.
        """
        d = {}
        for item_name, item in list(self.__item__.items.items()) \
                + [(u'global', self.__item__)]:
            for var, desc in item.var_info():
                if var not in d:
                    d[var] = {u'source': []}
                d[var][u'source'].append(item_name)
                d[var][u'value'] = None
                d[var][u'alive'] = False
        for var in self:
            val = self.get(var, _eval=False)
            if var not in d:
                # If a variable is not explicitly defined through var_info()
                # functions, and it's not a default-loggable type, then we
                # ignore it. This will for example ignore numpy arrays that
                # are defined in a script.
                if not self.is_default_loggable(var, val):
                    continue
                d[var] = {u'source': [u'?']}
            d[var][u'value'] = val
            d[var][u'alive'] = True
        return d

    def __reduce__(self):
        r"""Implements custom pickling. See var_store_pickle."""
        try:
            # If the parent var store is specified and is not None, then this
            # is a var store of an item, and we don't pickle it.
            if self.__parent__ is not None:
                return None
        except AttributeError:
            pass
        return (var_store_pickle, (self.inspect(), ))


class VarStorePickle(VarStore):

    r"""A read-only view of the var_store, which is used to inspect the Python
    workspace in the debug window and variable inspector. The real var_store is
    pickled using the `__reduce__()` method, transferred through the
    experiment's output_channel, and then unpickled into a `var_store_pickle`
    object.
    """
    def __init__(self, inspect):

        object.__setattr__(self, u'__inspect__', inspect)
        _vars = {}
        for var, info in inspect.items():
            if info[u'alive']:
                _vars[var] = info[u'value']
        object.__setattr__(self, u'__vars__', _vars)

    def __contains__(self, var):

        if var in self.__vars__:
            return True
        return False

    def __setattr__(self, var, val):

        raise RuntimeError('This var object is read-only')

    def get(self, var, default=None, _eval=True, valid=None):

        if var not in self:
            raise AttributeError(u'The variable %s does not exist' % var)
        return self.__vars__[var]

    def inspect(self):

        return self.__inspect__


class VarStoreIterator:

    r"""Implements an iterator over all variables in a var_store."""
    def __init__(self, var):

        self.var = var
        self.vars = var.vars()

    def __iter__(self):

        return self

    def __next__(self):

        # For Python 3
        return self.next()

    def next(self):

        if len(self.vars) == 0:
            raise StopIteration
        return self.vars.pop()


# Alias for backwards compatibility
var_store = VarStore
var_store_pickle = VarStorePickle
var_store_iterator = VarStoreIterator



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.misc import snake_case
from libopensesame.cistr import CIStr
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import ItemDoesNotExist
from libopensesame.item_stack import item_stack_singleton


# A list of names that cannot be used, because they cause trouble for one reason
# or another
INVALID_NAMES = [
    'response'  # causes conflict with correct_response variable
]


class ItemStore:
    """The `items` object provides dict-like access to the items. It's mainly
    useful for programmatically executing items.

    An `items` object is created automatically when the experiment starts.

    In addition to the functions listed below, the following semantics are
    supported:

    __Example__

    ~~~ .python
    # Programmatically prepare and run a sketchpad item.
    items.execute('my_sketchpad')
    # Check if an item exists
    if 'my_sketchpad' in items:
        print('my_sketchpad exists')
    # Delete an item
    del items['my_sketchpad']
    # Walk through all item names
    for item_name in items:
        print(item_name)
    ~~~

    [TOC]
    """
    built_in_types = ['sequence', 'loop', 'sketchpad', 'feedback',
                      'keyboard_response', 'mouse_response', 'sampler',
                      'synth', 'inline_script', 'logger']

    def __init__(self, experiment):
        r"""Constructor.

        Parameters
        ----------
        experiment : experiment.
            The experiment object.
        """
        self.__items__ = {}
        self.experiment = experiment

    def execute(self, name):
        r"""Executes the run and prepare phases of an item, and updates the
        item stack.

        Parameters
        ----------
        name : str
            An item name.

        Examples
        --------
        >>> items.execute('target_sketchpad')
        """
        self.prepare(name)
        self.run(name)

    def run(self, name):
        r"""Executes the run phase of an item, and updates the item stack.

        Parameters
        ----------
        name : str
            An item name.

        Examples
        --------
        >>> items.prepare('target_sketchpad')
        >>> items.run('target_sketchpad')
        """
        item_stack_singleton.push(name, 'run')
        self[name].run()
        item_stack_singleton.pop()

    def prepare(self, name):
        r"""Executes the prepare phase of an item, and updates the item stack.

        Parameters
        ----------
        name : str
            An item name.

        Examples
        --------
        >>> items.prepare('target_sketchpad')
        >>> items.run('target_sketchpad')
        """
        item_stack_singleton.push(name, 'prepare')
        self[name].prepare()
        item_stack_singleton.pop()

    def new(self, _type, name=None, script=None, allow_rename=True):
        r"""Creates a new item.

        Parameters
        ----------
        _type : unicode
            The item type.
        name : unicode, NoneType, optional
            The item name, or None to choose a unique name based on the item
            type.
        script : unicode, NoneType, optional
            A definition script, or None to start with a blank item.
        allow_rename : bool, optional
            Indicates whether OpenSesame can use a different name from the one
            that is provided as `name` to avoid duplicate names etc.

        Returns
        -------
        item
            The newly generated item.

        Examples
        --------
        >>> items.new('sketchpad', name='my_sketchpad')
        >>> items['my_sketchpad'].prepare()
        >>> items['my_sketchpad'].run()
        """
        oslogger.debug('creating %s' % _type)
        if allow_rename:
            name = self.valid_name(_type, suggestion=name)
        else:
            name = CIStr(name)
        if _type in self.experiment.plugin_manager:
            item = self.experiment.plugin_manager[_type].build(
                name, self.experiment, script)
            self.__items__[name] = item
            return item
        # Load one of the core items
        oslogger.debug(u"loading core item '%s' from '%s'" % (
            _type, self.experiment.module_container()))
        item_module = __import__(self.item_module(_type), fromlist=['dummy'])
        item_class = getattr(item_module, _type)
        item = item_class(name, self.experiment, script)
        self.__items__[name] = item
        return item

    def valid_name(self, item_type, suggestion=None):
        """Generates a unique name that is valid and resembles the desired
        name.

        Parameters
        ----------
        item_type : unicode
            The type of the item to suggest a name for.
        suggestion : unicode, NoneType, optional
            The desired name, or None to choose a name based on the item's
            type.

        Returns
        -------
        unicode
            A unique name.

        Examples
        --------
        >>> valid_name = items.valid_name('sketchpad', 'an invalid name')
        """
        if suggestion is None:
            name = f'new_{item_type}'
        else:
            name = self.experiment.syntax.sanitize(suggestion, strict=True,
                                                   allow_vars=False)
            # Empty names are not allowed, and we fall back to the item type
            if not name:
                name = item_type
            # Names cannot start with a number, and are therefore prefixed by
            # an underscore.
            elif name[0].isnumeric():
                name = '_' + name
        _name = name
        invalid_names = list(self) + INVALID_NAMES
        i = 1
        while _name in invalid_names:
            _name = f'new_{i}_{item_type}'
            i += 1
        return CIStr(_name)

    def _type(self, name):
        r"""Gets the type of an item.

        Parameters
        ----------
        name : unicode
            The name of an item.

        Returns
        -------
        unicode, NoneType
            The type of an item, or `None` if the item doesn't exist.

        Examples
        --------
        >>> print(items._type('target_sketchpad'))
        """
        if name not in self:
            return None
        return self[name].item_type
    
    def item_module(self, item_type):
        
        return f'{self.experiment.module_container()}.{snake_case(item_type)}'
        
    # The properties below emulate a dict interface.

    @property
    def __setitem__(self):
        return self.__items__.__setitem__

    @property
    def __delitem__(self):
        return self.__items__.__delitem__

    @property
    def __len__(self):
        return self.__items__.__len__

    @property
    def __iter__(self):
        return self.__items__.__iter__

    @property
    def items(self):
        return self.__items__.items

    @property
    def keys(self):
        return self.__items__.keys

    @property
    def values(self):
        return self.__items__.values

    @property
    def copy(self):
        return self.__items__.copy

    # The functions below are overridden to implement case insensitivity.

    def __contains__(self, name):

        if not isinstance(name, CIStr):
            try:
                name = CIStr(name)
            except AttributeError:
                return False
        return name in self.__items__

    def __getitem__(self, name):

        if not isinstance(name, CIStr):
            name = CIStr(name)
        try:
            return self.__items__[name]
        except KeyError:
            raise ItemDoesNotExist(name)


# Alias for backwards compatibility
item_store = ItemStore



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue
from libopensesame.base_response_item import BaseResponseItem
from openexp.keyboard import Keyboard


class KeyboardResponseMixin:

    r"""A mixin class that should be inherited along with base_response_item by
    all classes that want to collect keyboard responses.
    """
    def prepare_response_func(self):
        """See base_response_item."""
        self._keyboard = Keyboard(self.experiment, timeout=self._timeout,
                                  keylist=self._allowed_responses)
        if self.var.get(u'event_type', default=u'keypress') == u'keypress':
            return self._keyboard.get_key
        if self.var.event_type == u'keyrelease':
            return self._keyboard.get_key_release
        raise InvalidValue(f'event_type should be "keypress" or "keyrelease", '
                           f'not {self.var.event_type}')


class KeyboardResponse(KeyboardResponseMixin, BaseResponseItem):

    r"""An item for collecting keyboard responses."""
    description = u'Collects keyboard responses'
    process_feedback = True

    def reset(self):
        """See item."""
        self.var.flush = u'yes'
        self.var.timeout = u'infinite'
        self.var.duration = u'keypress'
        self.var.event_type = u'keypress'
        self.var.unset(u'allowed_responses')
        self.var.unset(u'correct_response')

    def prepare(self):
        """See item."""
        super().prepare()
        self._flush = self.var.flush == u'yes'

    def response_matches(self, test, ref):
        """See base_response_item."""
        return any(safe_decode(r) in self._keyboard.synonyms(test) for r in ref)

    def run(self):
        """See item."""
        if self._flush:
            self._keyboard.flush()
        super().run()

    def coroutine(self):
        """See coroutines plug-in."""
        self._keyboard.timeout = 0
        alive = True
        yield
        self._t0 = self.set_item_onset()
        if self._flush:
            self._keyboard.flush()
        while alive:
            key, time = self._keyboard.get_key()
            if key is not None:
                break
            alive = yield
        self.process_response((key, time))


# Alias for backwards compatibility
keyboard_response = KeyboardResponse
keyboard_response_mixin = KeyboardResponseMixin



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import BaseExtension


class RestoreUiElements(BaseExtension):

    r"""Restores key UI elements when they have been accidentally hidden."""
    @BaseExtension.as_thread(1000)
    def event_startup(self):

        if (
                self.main_window.toolbar_items.isVisible() and
                (self.menubar.isVisible() or self.toolbar.isVisible()) and
                self.main_window.dock_overview.isVisible()
        ):
            return
        oslogger.debug('some key ui elements are hidden')
        from libqtopensesame._input.confirmation import confirmation
        if not confirmation(
                self.main_window,
                _('Some important elements of the user interface have been hidden. Do you want to show them again?'),
                title=_('Restore user interface'),
                default='yes'
        ).show():
            return
        oslogger.debug('showing key ui elements')
        self.main_window.toolbar_items.setVisible(True)
        self.menubar.setVisible(True)
        self.toolbar.setVisible(True)
        self.main_window.dock_overview.setVisible(True)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets


class VariableInspectorCell(QtWidgets.QTableWidgetItem):
    r"""A cell for the variable-inspector table."""
    
    def __init__(self, text, info):
        r"""Constructor.

        Parameters
        ----------
        text : str
            The cell text.
        info : dict
            Variable info as returned by var_store.inspect().
        """
        a = QtCore.Qt.AlignLeft
        f = QtGui.QFont()
        if info[u'alive']:
            f.setBold(True)
        if not isinstance(text, str):
            try:
                len(text)
            except TypeError:
                if text is None:
                    text = u''
                elif isinstance(text, int) or isinstance(text, float):
                    text = str(text)
                    a = QtCore.Qt.AlignRight
                else:
                    text = safe_decode(text, errors=u'ignore')
            else:
                text = u','.join([safe_decode(s) for s in text])
        QtWidgets.QTableWidgetItem.__init__(self, text)
        self.setFont(f)
        self.setTextAlignment(a)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import re
from pyqode.core import api
from qtpy import QtWidgets
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pyqode_manager', category=u'extension')


class ConvertIndentationMode(api.Mode):
    """Comments/uncomments a set of lines using Ctrl+/."""
    
    def __init__(self):
        super().__init__()
        self._action_tabs_to_spaces = QtWidgets.QAction(
            _('Convert tabs to spaces'),
            self.editor
        )
        self._action_spaces_to_tabs = QtWidgets.QAction(
            _('Convert spaces to tabs'),
            self.editor
        )

    def on_state_changed(self, state):
        """
        Called when the mode is activated/deactivated
        """
        if state:
            self._action_spaces_to_tabs.triggered.connect(self._spaces_to_tabs)
            self.editor.add_action(
                self._action_spaces_to_tabs,
                sub_menu='Python'
            )
            self._action_tabs_to_spaces.triggered.connect(self._tabs_to_spaces)
            self.editor.add_action(
                self._action_tabs_to_spaces,
                sub_menu='Python'
            )
        else:
            self.editor.remove_action(
                self._action_spaces_to_tabs,
                sub_menu='Python'
            )
            self._action_spaces_to_tabs.triggered.disconnect(
                self._spaces_to_tabs
            )
            self.editor.remove_action(
                self._action_tabs_to_spaces,
                sub_menu='Python'
            )
            self._action_tabs_to_spaces.triggered.disconnect(
                self._tabs_to_spaces
            )

    def _spaces_to_tabs(self):

        code = self.editor.toPlainText()
        while True:
            match = re.search(
                u'((?<=\n){indent}+)|(\A{indent}+)'.format(
                    indent=u' ' * cfg.pyqode_tab_length
                ),
                code
            )
            if not match:
                break
            space_indent = u'\t' * \
                (len(match.group()) // cfg.pyqode_tab_length)
            code = code[:match.start()] + space_indent + code[match.end():]
        self.editor.setPlainText(code)
        self.editor.document().setModified(True)

    def _tabs_to_spaces(self):

        code = self.editor.toPlainText()
        while True:
            match = re.search(u'((?<=\n)\t+)|(\A\t+)', code)
            if not match:
                break
            space_indent = u' ' * len(match.group()) * cfg.pyqode_tab_length
            code = code[:match.start()] + space_indent + code[match.end():]
        self.editor.setPlainText(code)
        self.editor.document().setModified(True)



"""A simple consent form"""

category = "Form"
icon = "os-form_consent"
help_url = "manual/forms/readymade"
controls = [
    {
        "type": "line_edit",
        "var": "form_title",
        "label": "Form title",
        "tooltip": "Title to appear above the form text"
    },
    {
        "type": "line_edit",
        "var": "checkbox_text",
        "label": "Checkbox text",
        "tooltip": "Text for the checkbox"
    },
    {
        "type": "line_edit",
        "var": "accept_text",
        "label": "Accept-button text",
        "tooltip": "Text for the accept button"
    },
    {
        "type": "line_edit",
        "var": "decline_text",
        "label": "Decline-button text",
        "tooltip": "Text for the decline button"
    },
    {
        "type": "line_edit",
        "var": "decline_message",
        "label": "Message on decline",
        "tooltip": "A message shown when the participant declines"
    },
    {
        "type": "editor",
        "var": "form_text",
        "label": "Consent form text",
        "tooltip": "Text to display in the form body"
    }
]



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import numpy as np
from openexp._canvas import canvas
from libopensesame.oslogging import oslogger


class PsychoElement:

    r"""Together with Element, PsychoElement is the base object for all psycho
    sketchpad elements.
    """
    @property
    def win(self):
        return self._canvas.experiment.window

    def show(self):

        if self.visible:
            self._stim.draw()

    def _on_attribute_change(self, **kwargs):

        if self._canvas.auto_prepare:
            self.prepare()
            
    @staticmethod
    def _power_of_two(n):
        n = int(n)
        # A power of two always has exactly one bit set and does not have any
        # bitwise overlap with one number lower:
        # 4: 100
        # 3: 011
        # &: 000
        if (n & (n - 1) == 0) and n != 0:
            return n
        # The bit_length is the number of bits necessary to represent a number
        # so for 3 this would be 2: 11. The bit_length for one number lower in
        # this case is still 2: 10. The next power of two is 2 raised by this
        # number, so 2 ** 2 = 4.
        rounded_up = 2 ** ((n - 1).bit_length())
        oslogger.warning(
            f"Warning: {n} is not a power of two, rounding up to {rounded_up}.")
        return rounded_up

    def _mask(self, env, size, stdev):
        r"""Generates a PsychoPy mask for Gabor and NoisePatch stimuli.

        Parameters
        ----------
        env : str
            The envelope.
        size : int
            The stimulus size.
        stdev : int
            The standard deviation of the mask if the envelope is gaussian.

        Returns
        -------
        ndarray
            A PsychoPy mask, which is a numpy array.
        """
        # Create a PsychoPy mask
        env = canvas._match_env(env)
        if env == u'c':
            return u'circle', size
        if env == u'g':
            return u'gauss', 6 * stdev
        if env == u'r':
            return u'None', size
        if env == u'l':
            # Get the smallest power-of-two that is larger than or equal to 
            # the given size
            size = self._power_of_two(size)
            _env = np.zeros([size, size])
            for x in range(size):
                for y in range(size):
                    r = np.sqrt((x-size/2)**2+(y-size/2)**2)
                    _env[x, y] = (max(0, (0.5*size-r) / (0.5*size))-0.5)*2
            return _env, size
        raise ValueError('Invalid mask')


class RotatingElement:

    def _on_attribute_change(self, **kwargs):

        if u'rotation' in kwargs:
            self._stim.ori = kwargs.pop(u'rotation')
        if kwargs:
            super(RotatingElement, self)._on_attribute_change(**kwargs)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *


class ItemsAdapter:
    """Gives an editable view on (item_name, start_time, end_time, cond, 
    enabled) tuple (schedule) used by coroutines that functions like an
    (item_name, cond) tuple as used by sequences.
    """
    def __init__(self, schedule):

        self.schedule = schedule

    def __delitem__(self, index):

        del self.schedule[index]

    def __getitem__(self, index):

        item = self.schedule[index]
        return item[0], item[3], item[4]

    def __setitem__(self, index, val):

        self.schedule[index] = (val[0],) + \
            self.schedule[index][1:3] + (val[1],) + (val[2],)

    def __iter__(self):

        for item_name, start_time, end_time, cond, enabled in self.schedule:
            yield item_name, cond, enabled

    def __len__(self):

        return len(self.schedule)

    def insert(self, index, item):

        self.schedule.insert(
            index, ((item[0],) + (0, 0) + (item[1],) + (item[2],)))

    def append(self, item):

        self.schedule.append(
            ((item[0],) + (0, 0) + (item[1],) + (item[2],)))



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.widgets.tree_base_item import TreeBaseItem
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_unused_items_item', category=u'core')


class TreeUnusedItemsItem(TreeBaseItem):

    r"""Corresponds to the unused-items widget in the overview area."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        """
        super().__init__()
        self.setup(main_window)
        self.setIcon(0, self.theme.qicon(u'unused'))
        self.setToolTip(0, _(u'Unused items'))
        self._droppable = True
        self._draggable = False
        self.name = u'__unused__'
        i = 0
        for item_name in self.experiment.items.unused():
            self.experiment.items[item_name].build_item_tree(self, max_depth=1)
            i += 1
        self.setText(0, _(u'Unused items') + u' (%s)' % i)

    def droppable(self, data):

        return drag_and_drop.matches(data, [u'item-existing']) and \
            data[u'application-id'] == self.main_window._id()

    def drop_hint(self):

        return _(u'Move to unused items')

    def open_tab(self):

        self.main_window.tabwidget.open_unused()

    def ancestry(self):

        return u'__unused__', u'__unused__:0'

    def show_context_menu(self, pos):

        from libqtopensesame.widgets.unused_items_context_menu import \
            UnusedItemsContextMenu
        menu = UnusedItemsContextMenu(self.main_window, self)
        menu.popup(pos)


# Alias for backwards compatibility
tree_unused_items_item = TreeUnusedItemsItem



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import OSException
from ._base_task import BaseTask
from libopensesame.item_stack import item_stack_singleton


class ItemTask(BaseTask):
    """A task controls the coroutine for one item."""

    def __init__(self, coroutines, _item, start_time, end_time,
                 abort_on_end=False):
        if not hasattr(_item, u'coroutine'):
            raise OSException(f'{_item.item_type} not supported by coroutines')
        self._item = _item
        super().__init__(coroutines, start_time, end_time, abort_on_end)
        self.coroutines.event(u'initialize %s' % _item.coroutine)

    def step(self):
        item_stack_singleton.push(self._item.name, u'coroutines_step')
        retval = super().step()
        item_stack_singleton.pop()
        return retval

    def launch(self):
        item_stack_singleton.push(self._item.name, u'coroutines_prepare')
        self._item.prepare()
        item_stack_singleton.pop()
        # New-style coroutines take a coroutines keyword, which is used to
        # communicate the coroutines item. Old-style coroutines do not.
        try:
            self.coroutine = self._item.coroutine(coroutines=self.coroutines)
        except TypeError:
            self.coroutine = self._item.coroutine()
        self.coroutines.event('launch %s' % self._item)
        item_stack_singleton.push(self._item.name, u'coroutines_launch')
        self.coroutine.send(None)
        item_stack_singleton.pop()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue
from openexp.sampler import Sampler as OpenExpSampler
from libopensesame.item import Item
from libopensesame.base_response_item import BaseResponseItem
from libopensesame.keyboard_response import KeyboardResponseMixin
from libopensesame.mouse_response import MouseResponseMixin


class Sampler(BaseResponseItem, KeyboardResponseMixin, MouseResponseMixin):

    r"""An item for sound-file playback."""
    description = u'Plays a sound file in .wav or .ogg format'
    is_oneshot_coroutine = True

    def reset(self):
        """See item."""
        self.var.sample = u''
        self.var.pan = 0
        self.var.pitch = 1
        self.var.fade_in = 0
        self.var.stop_after = 0
        self.var.volume = 1.0
        self.var.duration = u'sound'

    def process_response(self, response_args):
        """See base_response_item."""
        if self.var.duration == u'mouseclick':
            MouseResponseMixin.process_response(self, response_args)
            return
        super().process_response(response_args)

    def prepare_response_func(self):
        """See base_response_item."""
        if isinstance(self.var.duration, (int, float)):
            return self._prepare_sleep_func(self.var.duration)
        if self.var.duration == u'keypress':
            return KeyboardResponseMixin.prepare_response_func(self)
        if self.var.duration == u'mouseclick':
            return MouseResponseMixin.prepare_response_func(self)
        if self.var.duration == u'sound':
            return lambda: None
        raise InvalidValue(f'Invalid duration: {self.var.duration}')

    def prepare(self):
        """See item."""
        super().prepare()
        if safe_decode(self.var.sample).strip() == u'':
            raise InvalidValue(f'No sample has been specified')
        sample = self.experiment.pool[self.var.sample]
        self.sampler = OpenExpSampler(self.experiment, sample)
        pan = self.var.pan
        if pan == -20:
            pan = u'left'
        elif pan == 20:
            pan = u'right'
        self.sampler.pan = pan
        self.sampler.volume = self.var.volume
        self.sampler.pitch = self.var.pitch
        self.sampler.fade_in = self.var.fade_in
        self.sampler.duration = self.var.stop_after
        self.sampler.block = self.var.duration == u'sound'

    def run(self):
        """See item."""
        self._t0 = self.set_item_onset()
        self.sampler.play()
        super().run()

    def coroutine(self):
        """See coroutines plug-in."""
        self.sampler.block = False
        yield
        self.set_item_onset()
        self.sampler.play()

    def var_info(self):
        """See item."""
        if self.var.get(u'duration', _eval=False, default=u'') in \
                [u'keypress', u'mouseclick']:
            return super().var_info()
        return Item.var_info(self)


# Alias for backwards compatibility
sampler = Sampler



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'font_widget', category=u'core')


class FontWidgetBase(BaseWidget):

    """A font selection widget"""
    max_size = 64
    font_list = [
        u'mono', u'sans', u'serif', u'arabic',
        u'chinese-japanese-korean', u'hebrew', u'hindi',
        _(u'other ...')
    ]

    font_changed = QtCore.Signal(['QString', int, bool, bool])

    def _apply(self):
        """Applies the controls."""
        self.family = self.ui.combobox_family.currentText()
        self.size = self.ui.spinbox_size.value()
        self.italic = self.ui.checkbox_italic.isChecked()
        self.bold = self.ui.checkbox_bold.isChecked()
        self.ui.label_example.setFont(self.get_font())
        self.font_changed.emit(
            self.family,
            self.size,
            self.italic,
            self.bold
        )

    def apply_family(self):
        """
        Applies the controls and optionally presents a full font selection
        dialog if the user has selected 'other ...'
        """
        if self.ui.combobox_family.currentText() == _(
                u'other ...',
                context=u'font_widget'
        ):
            font, ok = QtWidgets.QFontDialog.getFont(
                self.get_font(),
                parent=self._parent
            )
            if ok:
                self.family = str(font.family())
            else:
                self.family = self.experiment.var.get(
                    u'font_family',
                    _eval=False
                )
            self.update_family_combobox()
        self._apply()

    def get_font(self):
        """
        returns:
                type: QFont
        """
        font = QtGui.QFont(
            self.family,
            italic=self.italic,
            weight=QtGui.QFont.Bold if self.bold else QtGui.QFont.Normal
        )
        font.setPixelSize(min(self.max_size, self.size))
        return font

    def set_font(self, family=None, italic=None, bold=None, size=None):
        r"""Sets the widget to the specified font. For keywords, see
        initialize().
        """
        self.ui.combobox_family.activated.disconnect()
        self.ui.checkbox_italic.clicked.disconnect()
        self.ui.checkbox_bold.clicked.disconnect()
        self.ui.spinbox_size.editingFinished.disconnect()
        self.initialize(
            family=family,
            italic=italic,
            bold=bold,
            size=size,
            parent=self._parent
        )

    def initialize(
            self,
            experiment=None,
            family=None,
            italic=None,
            bold=None,
            size=None,
            parent=None
    ):
        r"""Initializes the widget.

        Parameters
        ----------
        experiment
            The experiment.
        family, optional
            The font family or None to use experiment default.
        italic, optional
            The font italic state or None to use experiment default.
        bold, optional
            font bold state or None to use experiment default.
        size, optional
            The font size or None to use experiment default.
        parent, optional
            A parent QWidget.
        """
        self._parent = parent if parent is not None else self.main_window
        if experiment is None:
            experiment = self.experiment
        self.family = experiment.var.get(u'font_family', _eval=False) \
            if family is None else family
        self.italic = experiment.var.get(u'font_italic', _eval=False) == u'yes' \
            if italic is None else italic
        self.bold = experiment.var.get(u'font_bold', _eval=False) == u'yes' \
            if bold is None else bold
        self.size = experiment.var.get(u'font_size', _eval=False) \
            if size is None else size
        if not isinstance(self.size, int) or self.size <= 0:
            self.size = 18
        if self.ui.combobox_family.findText(self.family) < 0:
            self.ui.combobox_family.addItem(self.family)
        self.ui.combobox_family.setCurrentIndex(
            self.ui.combobox_family.findText(self.family))
        self.ui.checkbox_italic.setChecked(self.italic)
        self.ui.checkbox_bold.setChecked(self.bold)
        self.ui.spinbox_size.setValue(self.size)
        self._apply()
        self.ui.combobox_family.activated.connect(self.apply_family)
        self.ui.checkbox_italic.clicked.connect(self._apply)
        self.ui.checkbox_bold.clicked.connect(self._apply)
        self.ui.spinbox_size.editingFinished.connect(self._apply)

    def update_family_combobox(self):
        """Updates the family combobox to include a custom font."""
        self.ui.combobox_family.activated.disconnect()
        self.ui.combobox_family.clear()
        l = self.font_list[:]
        if self.family not in l:
            l += [self.family]
        self.ui.combobox_family.insertItems(0, l)
        self.ui.combobox_family.setCurrentIndex(
            self.ui.combobox_family.findText(self.family))
        self.ui.combobox_family.activated.connect(self.apply_family)


class FontWidgetHorizontal(FontWidgetBase):

    def __init__(self, parent):
        """
        Constructor.

        Arguments:
        experiment	--	The experiment.

        Keywords arguments:
        parent		--	The parent QWidget. (default=None)
        """
        super().__init__(parent, ui=u'widgets.font_widget_horizontal')


class FontWidget(FontWidgetBase):

    def __init__(self, parent):
        """
        Constructor.

        Arguments:
        experiment	--	The experiment.

        Keywords arguments:
        parent		--	The parent QWidget. (default=None)
        """
        super().__init__(parent, ui=u'widgets.font_widget')


# Alias for backwards compatibility
font_widget = FontWidget
font_widget_base = FontWidgetBase
font_widget_horizontal = FontWidgetHorizontal



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.widgets.tree_append_menu import TreeAppendMenu
from libqtopensesame.misc.translate import translation_context
_ = translation_context('item_context_men', category='core')


class ItemContextMenu(BaseSubcomponent, QtWidgets.QMenu):
    """Provides a basic context menu for an item.
    
    Parameters
    ----------
    main_window : QtOpenSesame
    treeitem : TreeItemItem
    """
    
    def __init__(self, main_window, treeitem):
        super().__init__(main_window)
        self.setup(main_window)
        self.treeitem = treeitem
        self.addAction(self.theme.qicon(self.item.item_icon()), _('Open'),
                       self.item.open_tab)
        self.addSeparator()
        self.add_action(
            "accessories-text-editor", _("Rename"), self.treeitem.start_rename,
            cfg.shortcut_rename)
        if not self.treewidget.overview_mode and \
                self.treeitem.parent() is not None:
            self.add_action("accessories-text-editor",
                            _("Edit run-if statement"),
                            self.treeitem.start_edit_runif,
                            cfg.shortcut_edit_runif)
        self.addSeparator()
        self.add_action("edit-copy", _("Copy (unlinked)"),
                        self.treeitem.copy_unlinked,
                        cfg.shortcut_copy_clipboard_unlinked)
        self.add_action("edit-copy", _("Copy (linked)"),
                        self.treeitem.copy_linked,
                        cfg.shortcut_copy_clipboard_linked)
        if self.treeitem.item.children():
            self.add_action("edit-copy", _("Copy (shallow)"),
                            self.treeitem.copy_shallow)
        if self.treeitem.clipboard_data() is not None:
            self.add_action("edit-paste", _("Paste"),
                            self.treeitem.paste,
                            cfg.shortcut_paste_clipboard)
        if self.treeitem.is_deletable():
            self.addSeparator()
            self.add_action("list-remove", _("Delete"),
                            self.treeitem.delete, cfg.shortcut_delete)
            self.add_action("list-remove",
                            _("Permanently delete all linked copies"),
                            self.treeitem.permanently_delete,
                            cfg.shortcut_permanently_delete)
        elif self.treeitem.is_unused():
            self.addSeparator()
            self.add_action("list-remove", _("Permanently delete"),
                            self.treeitem.permanently_delete,
                            cfg.shortcut_permanently_delete)
        if self.treeitem.has_append_menu():
            # An append menu for sequence items
            menu = TreeAppendMenu(self.treeitem.treeWidget(), self.treeitem)
            action = QtWidgets.QAction(self.theme.qicon('list-add'),
                                       'Append item', self)
            action.setMenu(menu)
            self.addSeparator()
            self.addAction(action)
        self.addSeparator()
        if self.treeitem.isDisabled():
            self.add_action('media-playback-pause', _('Enable'),
                            self.treeitem.enable)
        else:
            self.add_action('media-playback-pause', _('Disable'),
                            self.treeitem.disable)
        self.addSeparator()
        self.add_action("help", _("Help"), self.item.open_help_tab)

    def add_action(self, icon, text, func, shortcut=None):
        """A convenience function for adding menu actions.

        Parameters
        ----------
        icon
            An icon name.
        text
            A menu text.
        func
            A function to call when the action is activated.
        shortcut, optional
            A key sequence to activate the action.

        Returns
        -------
        QAction
        """
        action = self.addAction(self.theme.qicon(icon), text, func)
        if shortcut is not None:
            action.setShortcut(QtGui.QKeySequence(shortcut))
            action.setShortcutContext(QtCore.Qt.WidgetShortcut)
        return action

    @property
    def item(self):
        return self.treeitem.item

    @property
    def treewidget(self):
        return self.treeitem.treeWidget()


# Alias for backwards compatibility
item_context_menu = ItemContextMenu



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import FStringError, FStringSyntaxError
from libopensesame.item_stack import item_stack_singleton
import warnings


class BasePythonWorkspace:
    """Provides a basic Python workspace for use in the GUI. This avoids
    unnecessarily importing the entire runtime API.
    
    Parameters
    ----------
    experiment : Experiment
    """
    def __init__(self, experiment):
        self.experiment = experiment
        
    @property
    def _globals(self):
        return self.experiment.var.__vars__

    def check_syntax(self, script):
        """Checks whether a Python script is syntactically correct.

        Parameters
        ----------
        script : unicode
            A Python script.

        Returns
        -------
        int
            0 if script is correct, 1 if there is a syntax warning, and 2 if
            there is a syntax error.
        """
        with warnings.catch_warnings(record=True) as warning_list:
            try:
                self._compile(safe_decode(script))
            except Exception as e:
                return 2
        if warning_list:
            return 1
        return 0

    def run_file(self, path):
        """Reads and executes a files.

        Parameters
        ----------
        path : str
            The full path to a Python file.
        """
        with safe_open(path) as fd:
            script = fd.read()
        bytecode = self._compile(script)
        self._exec(bytecode)

    def _compile(self, script):
        """Compiles a script into bytecode.

        Parameters
        ----------
        script : unicode
            A Python script.

        Returns
        -------
        code
            The compiled script.
        """
        return compile(script, '<string>', 'exec')  # __ignore_traceback__

    def _exec(self, bytecode):
        """Executes bytecode.

        Parameters
        ----------
        bytecode : code
            A chunk of bytecode.
        """
        exec(bytecode, self._globals)  # __ignore_traceback__

    def _eval(self, bytecode):
        """Evaluates bytecode.

        Parameters
        ----------
        bytecode : code
            A chunk of bytecode.

        Returns
        -------
        The evaluated value of the bytecode
        """
        return eval(bytecode, self._globals)  # __ignore_traceback__
        
    def eval_fstring(self, fs, include_local=False):
        """Evaluates an f-string.

        Parameters
        ----------
        fs : str
            An f-string
        include_local : bool, optional
            If True, the variable store of the current item is merged into the
            Python workspace. This allows items to evaluate f-strings that
            take into account the item's local variables.

        Returns
        -------
        A string corresponding to the evaluated f-string.
        """
        if include_local:
            item_name, phase = item_stack_singleton[-1]
            _globals = self._globals.copy()
            _globals.update(self.experiment.items[item_name].var.__vars__)
        else:
            _globals = self._globals
        fs_escaped = fs.replace(r"'''", r"\'\'\'")
        try:
            return eval(f"f'''{fs_escaped}'''", _globals)  # __ignore_traceback__
        except SyntaxError:
            raise FStringSyntaxError(
                f'The following text contains invalid f-string expression:\n\n~~~ .text\n{fs}\n~~~\n\n')
        except Exception:
            raise FStringError(
                f'Failed to evaluate f-string expression in the following text:\n\n~~~ .text\n{fs}\n~~~\n\n')

    # The properties below emulate a dict interface.

    @property
    def __setitem__(self):
        return self._globals.__setitem__

    @property
    def __delitem__(self):
        return self._globals.__delitem__

    @property
    def __getitem__(self):
        return self._globals.__getitem__

    @property
    def __len__(self):
        return self._globals.__len__

    @property
    def __iter__(self):
        return self._globals.__iter__

    @property
    def items(self):
        return self._globals.items

    @property
    def keys(self):
        return self._globals.keys

    @property
    def values(self):
        return self._globals.values

    @property
    def copy(self):
        return self._globals.copy


# Alias for backwards compatibility
base_python_workspace = BasePythonWorkspace



"""Adds a Help menu"""

settings = {
    "online_help_base_url": {
        "default": "http://osdoc.cogsci.nl",
        "ide": "http://rapunzel.cogsci.nl"
    },
    "online_help_sitemap": {
        "default": "http://osdoc.cogsci.nl/[version]/sitemap.yml",
        "ide": "http://rapunzel.cogsci.nl/sitemap.yml"
    },
    "online_help_psychopy": {
        "default": True,
        "ide": False
    }
}
modes = ["ide", "default"]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._keyboard.keyboard import Keyboard
from psychopy import event
import pyglet.window.key
from openexp.backend import configurable


class Psycho(Keyboard):

    r"""This is a keyboard backend built on top of PsychoPy. For function
    specifications and docstrings, see `openexp._keyboard.keyboard`.
    """
    # The keymap is an incomplete attempt at translating keys from the PyGame
    # names to the names used by PsychoPy
    keymap = {
        '!': 'exclamation',
        '"': 'doublequote',
        '#': 'hash',
        '$': 'dollar',
        '&': 'ampersand',
        '\'': 'quoteleft',
        '(': None,
        ')': None,
        '*': 'asterisk',
        '+': 'plus',
        ',': 'comma',
        '-': 'minus',
        '.': None,
        '/': 'slash',
        ':': 'colin',
        ';': 'semicolon',
        '=': 'equal',
        '>': 'greater',
        '?': 'question',
        '@': 'at',
        '[': 'bracketleft',
        '\\': 'backslash',
        ']': 'bracketright',
        '^': None,
        '_': 'underscore'
    }

    def __init__(self, experiment, **resp_args):

        Keyboard.__init__(self, experiment, **resp_args)
        self._latest_modifiers = {}

    def valid_keys(self):

        l = []
        for i in dir(pyglet.window.key):
            if isinstance(getattr(pyglet.window.key, i), int):
                l.append(i)
        return l

    @configurable
    def get_key(self):

        keylist = self.keylist
        if keylist is not None:
            keylist += [u'escape']
        timeout = self.timeout
        start_time = self.experiment.clock.time()
        while True:
            # Don't use the timeStamped keyword, because it bugs on older
            # versions of PsychoPy (at least 1.79.01 under Ubuntu 14.04).
            # See https://github.com/smathot/OpenSesame/issues/381
            time = self.experiment.clock.time()
            keys = event.getKeys(keylist, modifiers=True)
            for key, modifiers in keys:
                if key == u'escape':
                    self.experiment.pause()
                if keylist is None or key in keylist:
                    self._latest_modifiers = modifiers
                    return key, time
            if timeout is not None and time - start_time >= timeout:
                break
        self._latest_modifiers = {}
        return None, time

    def get_mods(self):

        return [
            m
            for m in (u'shift', u'ctrl', u'alt')
            if self._latest_modifiers.get(m, False)
        ]

    def synonyms(self, key):

        if key is None:
            return [None, 'None', 'none', 'NONE']

        # Respond correctly if a keycode is passed, rather than a Unicode string
        # key description.
        if type(key) == int:
            l = [key, pyglet.window.key.symbol_string(key).lower()]
            if l[-1].upper() != l[-1].lower():
                l.append(l[-1].upper())
            return l

        # Sanity check
        if not isinstance(key, str):
            raise InvalidKeyName(key)

        # Make a list of all conceivable ways that a key might be referred to.
        l = [key]
        if key != key.upper():
            l.append(key.upper())
        if key != key.lower():
            l.append(key.lower())
        for char, name in self.keymap.items():
            if key == char:
                l.append(name)
            if key.lower() == name:
                l.append(char)
        # Make sure that we can deal with None/ timeout responses
        if key.lower() == u'none':
            l.append(None)
        # Make sure that we convert numeric strings to ints as well
        try:
            i = int(key)
            l.append(i)
        except:
            pass
        return l

    def flush(self):

        keypressed = False
        for key in event.getKeys():
            if key == u"escape":
                self.experiment.pause()
            keypressed = True
        return keypressed

    def _keycode_to_str(self, keycode):

        return pyglet.window.key.symbol_string(keycode).lower()


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



# -*- coding: utf-8 -*-
"""
@author: Daniel Schreij

This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

This module is distributed under the Apache v2.0 License.
You should have received a copy of the Apache v2.0 License
along with this module. If not, see <http://www.apache.org/licenses/>.
"""
import QNotifications
import time
import platform
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension


__author__ = u"Daniel Schreij"
__license__ = u"GPLv3"


class Notifications(BaseExtension):

    def event_startup(self):

        self.old_notifications = {}
        self.expiration_time = 60  # a minute
        self._suspend = False
        self.notification_area = QNotifications.QNotificationArea(
            self.tabwidget,
            useGlobalCSS=True
        )
        self.notification_area.move(0, 15)
        self.notification_area.setEntryEffect(u'fadeIn', 200)
        self.notification_area.setExitEffect(u'fadeOut', 200)

    def event_notify(
            self,
            message,
            category='primary',
            timeout=5000,
            always_show=False,
            buttontext=None
    ):
        """ Show a notification 'message' in the style 'notification type' for
        'timeout' milliseconds (where 0 milliseconds displays the notification
        indefinitely, until the user removes it)."""
        if self._suspend:
            return
        current_time = time.time()
        # See if notification has been shown before. If it is within
        # self.expiration_time, don't show it again.
        if not always_show:
            if (message, category, timeout, buttontext) in self.old_notifications:
                prev_time = self.old_notifications[(
                    message,
                    category,
                    timeout,
                    buttontext
                )]
                if current_time - prev_time < self.expiration_time:
                    return
            # Add notification to old notifications list.
            self.old_notifications[(
                message,
                category,
                timeout,
                buttontext
            )] = current_time
        try:
            self.notification_area.display(
                message, category, timeout, timeout != 0, buttontext
            )
        except TypeError:
            # Older versions of QNotifications did not support the autohide
            # option
            self.notification_area.display(
                message, category, timeout, buttontext
            )

    def event_notify_suspend(self):

        self._suspend = True

    def event_notify_resume(self):

        self._suspend = False



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy import QtWidgets, QtGui
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc.config import cfg
_ = translation_context(u'tree_append_menu', category=u'core')


class AppendExistingAction(BaseSubcomponent, QtWidgets.QAction):
    r"""An action for appending existing items, either as linked or unlinked
    copies.
    """
    
    def __init__(self, append_menu, menu, item_name):
        r"""Constructor.

        Parameters
        ----------
        append_menu : tree_append_menu
            The main menu.
        menu : QMenu
            The submenu for this action.
        item_name : unicode
            The name of the item to be appended.
        """
        super().__init__(menu)
        self.setup(append_menu)
        self.setText(item_name)
        self.item_name = item_name
        self.setIcon(
            self.theme.qicon(self.experiment.items[item_name].item_icon()))

    def append_item(self, target_item_name):
        r"""Performs the append operation.

        Parameters
        ----------
        target_item_name : unicode
            The name of the item (normally a sequence) to which the append
            should be applied.
        """
        target_item = self.experiment.items[target_item_name]
        item_name = self.item_name
        target_item.insert_child_item(item_name, len(target_item.items))


class AppendNewAction(BaseSubcomponent, QtWidgets.QAction):
    """An action for appending a new item.
    
    Parameters
    ----------
    append_menu : tree_append_menu
        The main menu.
    menu : QMenu
        The submenu for this action.
    item_type : unicode
        The type of item to be created and appended.
    """
    
    def __init__(self, append_menu, menu, item_type):
        super().__init__(menu)
        self.setup(append_menu)
        self.item_type = item_type
        self.setText(item_type)
        if item_type not in self.experiment.core_items:
            icon = self.theme.qicon(self.plugin_manager[item_type].icon)
        else:
            icon = self.theme.qicon(item_type)
        self.setIcon(icon)

    def append_item(self, target_item_name):
        r"""Performs the append operation.

        Parameters
        ----------
        target_item_name : unicode
            The name of the item (normally a sequence) to which the append
            should be applied.
        """
        target_item = self.experiment.items[target_item_name]
        item = self.experiment.items.new(
            self.item_type, catch_exceptions=False)
        if item is not None:
            target_item.insert_child_item(item.name, len(target_item.items))


class TreeAppendMenu(BaseSubcomponent, QtWidgets.QMenu):
    """An append item menu.
    
    Parameters
    ----------
    tree_overview : tree_overview
        The tree_overview with which this menu is associated.
    target_treeitem : tree_item_item, optional
        The treewidget item which corresponds to the sequence to which the
        append operation should be applied. If `None`, the top-level item
        from the tree_overview is used.
    """
    
    def __init__(self, tree_overview, target_treeitem=None):
        super().__init__(tree_overview)
        self.setup(tree_overview)
        self.target_treeitem = target_treeitem
        self.tree_overview = tree_overview
        self.action_new_items = QtWidgets.QAction(
            self.theme.qicon(u'list-add'), _(u'Append new item'), self)
        self.addAction(self.action_new_items)
        self.action_linked_copy = QtWidgets.QAction(
            self.theme.qicon(u'edit-copy'),
            _(u'Append existing item (linked)'), self)
        self.addAction(self.action_linked_copy)
        self.aboutToShow.connect(self.refresh)
        self.triggered.connect(self.append_item)
        self._new_items_menu = None
        self.extension_manager.register_extension(self)

    def append_item(self, action):
        r"""Performs the append action.

        Parameters
        ----------
        action : QAction
            A append_existing_action or append_new_action.
        """
        target_item_name, target_item_ancestry = self.target_treeitem.ancestry()
        action.append_item(target_item_name)
        self.tree_overview.structure_change.emit()
        if self.tree_overview.overview_mode:
            # In overview mode, we need to explicitly update the target item,
            # because this is not linked to the structure_change signal.
            self.experiment.items[target_item_name].update()

    def refresh(self):
        r"""Refreshes the menu before it is shown. This is necessary because
        the structure of the experiment may change.
        """
        self._items = None
        if self.target_treeitem is None:
            self.target_treeitem = self.tree_overview.topLevelItem(0)
        self.action_linked_copy.setMenu(self.existing_items_menu())
        self.action_new_items.setMenu(self.new_items_menu())

    def existing_items_menu(self):
        r"""Generates a menu with existing items.

        Parameters
        ----------
        linked : bool, optional
            Indicates whether the actions should result in linked (True) or
            unliked (False) copies.

        Returns
        -------
        QMenu
        """
        self._items = []
        target_item_name = self.target_treeitem.item.name
        for item_name in self.experiment.items:
            if target_item_name in \
                    self.experiment.items[item_name].children() or \
                    item_name == target_item_name:
                continue
            self._items.append(item_name)
        self._items.sort()
        m = QtWidgets.QMenu(self)
        for item_name in self._items:
            m.addAction(AppendExistingAction(self, m, item_name))
        return m

    def new_items_menu(self):
        r"""Generates a menu with new items.

        Returns
        -------
        QMenu
        """
        if self._new_items_menu is not None:
            return self._new_items_menu
        self._new_items_menu = QtWidgets.QMenu(self)
        for item_type in self.experiment.core_items + [None] + \
                [plugin.name for plugin in self.plugin_manager]:
            # Ignoring disabled extensions
            cfg_key = f'plugin_enabled_{item_type}'
            if cfg_key in cfg and not cfg[cfg_key]:
                continue
            # Ignoring unsupported items
            if not self.experiment.items.is_supported(item_type):
                continue
            if item_type is None:
                self._new_items_menu.addSeparator()
            else:
                self._new_items_menu.addAction(
                    AppendNewAction(self, self._new_items_menu, item_type))
        return self._new_items_menu

    def event_change_experiment(self): self._new_items_menu = None
    def event_open_experiment(self, path): self._new_items_menu = None
    def event_startup(self): self._new_items_menu = None


# Alias for backwards compatibility
tree_append_menu = TreeAppendMenu



"""Shows a list of example experiments"""

icon = "help-contents"
label = "Example experiments"
menu = {
    "index": -1,
    "separator_after": False,
    "separator_before": True,
    "submenu": "Tools"
}
tooltip = "Shows a list of example experiments"



"""Implements undo/ redo"""

icon = "edit-redo"
label = "Redo"
shortcut = "Alt+Ctrl+Shift+Z"
tooltip = "Redo most recently undone action"
menu = {
    "index": 0,
    "separator_after": True,
    "submenu": "Edit"
}
toolbar = {
    "index": -1,
    "separator_before": True
}
settings = {
    "max_undo_history": 100
}



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from openexp._color.color import Color
from openexp._canvas._richtext.richtext import RichText
from openexp._canvas.canvas import Canvas
from openexp._coordinates.coordinates import Coordinates
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.config import cfg
from qtpy import QtWidgets, QtGui, QtCore
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sketchpad', category=u'item')


class QtCanvas(Canvas, Coordinates):

    r"""A dummy Canvas that can be passed to the RichText element."""
    def __init__(self, experiment):

        Canvas.__init__(self, experiment)
        Coordinates.__init__(self)


class QtRichText(RichText):

    r"""Disables the pyqt initialization in the RichText element, because it
    is not necessary in the context of the GUI.
    """
    def _init_pyqt(self, exp):

        pass

    def _register_font(self, exp, font, fd=None):

        RichText._register_font(self, exp, font, QtGui.QFontDatabase())


class SketchpadCanvas(QtWidgets.QGraphicsScene):

    r"""A partial implementation of a canvas, so that sketchpad elements can
    draw to the canvas just as in runtime.
    """
    def __init__(self, sketchpad):
        r"""constructor.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        """
        self.sketchpad = sketchpad
        self.background_color = sketchpad.var.background
        self.placeholder_color = cfg.sketchpad_placeholder_color
        self.grid = 32
        super().__init__(self.sketchpad.main_window)
        self.indicator = None

    @property
    def selected_element_tool(self):

        return self.sketchpad.sketchpad_widget.selected_element_tool

    def dragMoveEvent(self, e):
        r"""Accepts drag-moves to implement moving items.

        Parameters
        ----------
        e : QGraphicsSceneDragDropEvent
        """
        data = drag_and_drop.receive(e)
        if data[u'type'] != u'sketchpad-element-move':
            e.ignore()
            return
        ex, ey = self.cursor_pos(e, grid=True)
        dx = ex-self.start_move_pos[0]
        dy = ey-self.start_move_pos[1]
        self.start_move_pos = ex, ey
        if abs(dx) < self.grid and abs(dy) < self.grid:
            e.accept()
            return
        self.sketchpad.move_elements(self.sketchpad.selected_elements(),
                                     dx=dx, dy=dy)
        self.sketchpad.draw()
        e.accept()

    def wheelEvent(self, e):
        r"""Scrolls the graphics view based on a wheel event (i.e. mouse-scroll
        event.

        Parameters
        ----------
        e : QGraphicsViewWheelEvent
        """
        if not QtCore.Qt.ControlModifier & e.modifiers():
            return
        self.sketchpad.zoom_diff(e.delta())

    def mouseDoubleClickEvent(self, e):
        r"""Processes mouse-press events, to handle element selection.

        Parameters
        ----------
        e : QGraphicsSceneMouseEvent
        """
        if not QtCore.Qt.LeftButton & e.button():
            # Only accept right clicks
            return
        # Select the pointed-at element
        element = self.element_at(e.scenePos())
        if element is None:
            return
        element.show_edit_dialog()

    def mousePressEvent(self, e):
        """
        desc:
                Processes mouse-press events. Mouse presses can have a number of
                effects.

                - All selected elements are deselected unless Control is pressed
                - Left clicks:
                        - If the pointer tool is selected, clicking an element selects
                          it.
                        - If an element tool is selected, clicking gives the element
                          tool the opporunity (to start a) drawing operation.
                - Right clicks:
                        - If an element is selected, a context menu is shown.

        arguments:
                e:
                        type:	QMouseEvent
        """
        # When control is not pressed, we deselect all currently selected
        # elements first
        if not QtCore.Qt.ControlModifier & e.modifiers():
            for element in self.sketchpad.elements:
                element.select(False)
        element = self.element_at(e.scenePos())
        if QtCore.Qt.RightButton & e.button() and element is not None:
            # The right buttons pops up a context menu.
            element.show_context_menu(e.screenPos())
        if QtCore.Qt.LeftButton & e.button():
            # The left button selects and drags.
            if self.selected_element_tool is not None:
                # When in pointer-tool mode, mouse clicks create new elements.
                self.sketchpad.add_element(
                    self.selected_element_tool.mouse_press(self.sketchpad,
                                                           self.cursor_pos(e)))
            elif element is None:
                self.sketchpad.select_pointer_tool()
            else:
                element.select()
                self.sketchpad.show_element_settings(element)
                x, y = self.cursor_pos(e, grid=False)
                self.start_move_pos = x, y
                data = {
                    u'type': u'sketchpad-element-move',
                    u'from_x': x,
                    u'from_y': y
                }
                drag_and_drop.send(e.widget(), data)

    def mouseReleaseEvent(self, e):
        r"""Mouse-release events give elements the opportunity to finish a
        drawing operation.

        Parameters
        ----------
        e : QMouseEvent
        """
        if self.selected_element_tool is not None:
            self.sketchpad.add_element(
                self.selected_element_tool.mouse_release(self.sketchpad,
                                                         self.cursor_pos(e)))

    def mouseMoveEvent(self, e):
        r"""Processes mouse-move events, to handle highlighting and cursor-
        position display.

        Parameters
        ----------
        e : QMouseEvent
        """
        cursor_pos = self.cursor_pos(e)
        self.sketchpad.set_cursor_pos(cursor_pos)
        for element in self.sketchpad.elements:
            element.highlight(False)
        # Only highlight elements if the pointer tool is selected
        if self.selected_element_tool is None:
            element = self.element_at(e.scenePos())
            if element is not None:
                element.highlight()
        else:
            self.selected_element_tool.mouse_move(self.sketchpad, cursor_pos)

    def keyPressEvent(self, e):
        r"""Processes key-press events, to handle deletion, movement, etc.

        Parameters
        ----------
        e : QKeyEvent
        """
        if e.key() == QtCore.Qt.Key_Delete:
            self.sketchpad.remove_elements(self.sketchpad.selected_elements())
        elif e.key() == QtCore.Qt.Key_Up:
            self.sketchpad.move_elements(self.sketchpad.selected_elements(),
                                         dy=-self.grid)
        elif e.key() == QtCore.Qt.Key_Down:
            self.sketchpad.move_elements(self.sketchpad.selected_elements(),
                                         dy=self.grid)
        elif e.key() == QtCore.Qt.Key_Left:
            self.sketchpad.move_elements(self.sketchpad.selected_elements(),
                                         dx=-self.grid)
        elif e.key() == QtCore.Qt.Key_Right:
            self.sketchpad.move_elements(self.sketchpad.selected_elements(),
                                         dx=self.grid)
        else:
            super(sketchpad_canvas, self).keyPressEvent(e)
            return
        self.sketchpad.draw()

    def element_at(self, pos):
        r"""Gets the element at a specific position.

        Parameters
        ----------
        pos
            An (x,y) tuple.

        Returns
        -------
        A sketchpad element (base_element) or None of no element was at pos.
        """
        # First try to see if there's an exact position match ...
        graphics_item = self.itemAt(pos, QtGui.QTransform())
        # ... if not, try if there's an element that encompasses the position.
        if graphics_item is None:
            for item in self.items():
                if item.boundingRect().contains(pos):
                    graphics_item = item
        # ... else don't find an element.
        if graphics_item is None:
            return
        # If the item is part of a group, we want the group, not the invdividual
        # item, because the group has the element property.
        if graphics_item.group() is not None:
            graphics_item = graphics_item.group()
        try:
            return graphics_item.element
        except AttributeError:  # Appears to happen in rare race conditions
            return

    def cursor_pos(self, e, grid=True):
        r"""Gets the position of the mouse cursor.

        Parameters
        ----------
        e : QMouseEvent
        grid : bool, optional
            Indicates whether the cursor should be locked to the grid.

        Returns
        -------
        An (x, y) tuple with the the coordinates of the mouse cursor.
        """
        pos = e.scenePos().toPoint()
        if grid:
            x = pos.x() + 0.5 * self.grid
            y = pos.y() + 0.5 * self.grid
            x = x - x % self.grid
            y = y - y % self.grid
        else:
            x = pos.x()
            y = pos.y()
        return x, y

    def notify(self, msg):
        r"""Adds a notification message."""
        self.notifications.append(msg)

    def is_var(self, val):
        r"""Determines whether a value is a string containing variable
        references (i.e. "[my_var]").

        Parameters
        ----------
        val
            A value.

        Returns
        -------
        True if val contains variable references, False otherwise.
        """
        if not isinstance(val, str):
            return False
        return self.sketchpad.syntax.contains_variables(val)

    def _pixmap(self, fname):
        r"""Safely returns a QPixmap.

        Returns
        -------
        A QPixmap object.
        """
        if not os.path.exists(fname):
            self.notify(
                _(u'Image name "%s" is unknown or variably defined, using fallback image')
                % fname)
            return self.sketchpad.theme.qpixmap(u'dialog-question')
        return QtGui.QPixmap(fname)

    def _pen(self, color, penwidth, alpha=255):

        color = self._color(color)
        color.setAlpha(alpha)
        p = QtGui.QPen(color)
        p.setWidth(self._penwidth(penwidth))
        return p

    def _brush(self, color):
        r"""Safely returns a QBrush.

        Returns
        -------
        A QBrush object.
        """
        return QtGui.QBrush(self._color(color))

    def _penwidth(self, penwidth):
        r"""Safely returns a penwidth.

        Returns
        -------
        An int penwidth.
        """
        if type(penwidth) not in (int, float):
            self.notify(
                _(u'Penwidth "%s" is unknown or variably defined, using 1') % penwidth)
            return 1
        return penwidth

    def _scale(self, scale):
        r"""Safely returns a scale.

        Returns
        -------
        An float scale.
        """
        if not isinstance(scale, (int, float)):
            self.notify(
                _(u'Scale "%s" is unknown or variably defined, using 1')
                % scale
            )
            return 1
        return scale

    def _rotation(self, rotation):
        r"""Safely returns a rotation.

        Returns
        -------
        A rotation.
        """
        if rotation is None:
            return 0
        if not isinstance(rotation, (int, float)):
            self.notify(
                _('Rotation "%s" is unknown or variably defined, using 0')
                % rotation
            )
            return 0
        return rotation

    def _point(self, i, x, y, center, scale):
        r"""Safely returns a QPoint.

        Parameters
        ----------
        i
            A QGraphicsItem.
        x
            An X coordinate.
        y
            A Y coordinate.
        center
            A boolean indicating whether the point should reflect the center of
            i (True), or the top-left (False).
        scale
            A scaling factor.

        Returns
        -------
        A QPoint object.
        """
        x = self._x(x)
        y = self._y(y)
        scale = self._scale(scale)
        if center:
            r = i.boundingRect()
            x -= r.width() * scale/2
            y -= r.height() * scale/2
        return QtCore.QPoint(x, y)

    def _color(self, _color):
        r"""Safely returns a QColor.

        Returns
        -------
        A QColor object.
        """
        if isinstance(_color, QtGui.QColor):
            return _color
        try:
            hexcolor = Color(self.sketchpad.experiment, _color).hexcolor
        except:
            self.notify(
                _(u'Color "%s" is unknown or variably defined, using placeholder color')
                % _color)
            return QtGui.QColor(self.placeholder_color)
        return QtGui.QColor(hexcolor)

    def _fill(self, fill):
        r"""Safely returns a fill value.

        Returns
        -------
        A fill value (bool).
        """
        if isinstance(fill, str):
            self.notify(
                _(u'Fill "%s" is unknown or variably defined, assuming filled')
                % fill)
            return True
        return bool(fill)

    def _x(self, x):
        r"""Safely returns an X coordinate.

        Returns
        -------
        An X coordinate.
        """
        if type(x) not in (int, float):
            self.notify(
                _('X coordinate "%s" is unknown or variably defined, using display center')
                % x)
            return 0
        return x

    def _y(self, y):
        r"""Safely returns a Y coordinate.

        Returns
        -------
        A Y coordinate.
        """
        if type(y) not in (int, float):
            self.notify(
                _('Y coordinate "%s" is unknown or variably defined, using display center')
                % y)
            return 0
        return y

    def _r(self, r):
        r"""Safely returns a radius

        Returns
        -------
        A radius.
        """
        if type(r) not in (int, float):
            self.notify(
                _('Radius "%s" is unknown or variably defined, using 50') % r)
            return 50
        return r

    def _p(self, r):
        r"""Safely returns a proportion

        Returns
        -------
        A radius.
        """
        if type(r) not in (int, float):
            self.notify(
                _('Proportion "%s" is unknown or variably defined, using .5')
                % r)
            return .5
        return r

    def _w(self, x):
        r"""Safely returns a width.

        Returns
        -------
        A width.
        """
        if type(x) not in (int, float):
            self.notify(
                _('Width "%s" is unknown or variably defined, using 100') % x)
            return 100
        return x

    def _h(self, y):
        r"""Safely returns a height.

        Returns
        -------
        A height.
        """
        if type(y) not in (int, float):
            self.notify(
                _('Height "%s" is unknown or variably defined, using 100') % y)
            return 100
        return y

    def _font_size(self, font_size):
        r"""Safely returns a font_size.

        Returns
        -------
        A font size.
        """
        if not isinstance(font_size, int) or font_size < 0:
            self.notify(_('Font size "%s" is invalid or variably defined, using 18')
                        % font_size)
            return 18
        return font_size

    def drawBackground(self, painter, rect):
        r"""Draws the background and the grid."""
        xc = int(self.xcenter())
        yc = int(self.ycenter())
        w = 2*xc
        h = 2*yc
        painter.fillRect(QtCore.QRect(-xc, -yc, w, h), self._color(
            self.background_color))
        painter.setPen(self._pen(cfg.sketchpad_grid_color,
                                 cfg.sketchpad_grid_thickness_thin,
                                 cfg.sketchpad_grid_opacity))
        painter.drawRect(QtCore.QRect(-xc, -yc, w, h))
        # Draw all lines except for the center ones, because they should be
        # thicker
        if self.grid > 1:
            for x in range(0, xc+1, self.grid):
                painter.drawLine(x, -yc, x, yc)
            for x in range(-self.grid, -xc-1, -self.grid):
                painter.drawLine(x, -yc, x, yc)
            for y in range(0, yc+1, self.grid):
                painter.drawLine(-xc, y, xc, y)
            for y in range(-self.grid, -yc-1, -self.grid):
                painter.drawLine(-xc, y, xc, y)
        # Draw thicker central lines
        painter.setPen(self._pen(cfg.sketchpad_grid_color,
                                 cfg.sketchpad_grid_thickness_thick,
                                 cfg.sketchpad_grid_opacity))
        painter.drawLine(-xc, 0, xc, 0)
        painter.drawLine(0, -yc, 0, yc)

    def clear(self):
        """Mimics canvas api. See openexp._canvas.canvas."""
        super(sketchpad_canvas, self).clear()
        self.notifications = []
        self.elements = []
        self.background_color = self.sketchpad.var.background

    def xcenter(self):
        """Mimics canvas api. See openexp._canvas.canvas."""
        return self.sketchpad.var.get(u'width') // 2

    def ycenter(self):
        """Mimics canvas api. See openexp._canvas.canvas."""
        return self.sketchpad.var.get(u'height') // 2

    def text(self, text, center=True, x=None, y=None, max_width=None,
             **properties):
        """Mimics canvas api. See openexp._canvas.canvas."""
        if u'color' in properties:
            properties[u'color'] = self._color(properties[u'color']).name()
        if u'font_size' in properties:
            properties[u'font_size'] = self._font_size(
                properties[u'font_size'])
        x = self._x(x)
        y = self._x(y)
        # We set the color backend to a dummy value. This avoids PyGame from
        # being unnecessarily loaded in the legacy color backend. Then we
        # restore the backend, also catching the case in which no color backend
        # was set.
        color_backend = (
            self.sketchpad.experiment.var.color_backend
            if u'color_backend' in self.sketchpad.experiment.var
            else None
        )
        self.sketchpad.experiment.var.color_backend = 'color'
        t = QtRichText(
            QtCanvas(self.sketchpad.experiment),
            text,
            x=x, y=y, center=center, max_width=max_width,
            **properties
        )
        # Restore the color backend so that we don't change the experiment
        if color_backend is None:
            del self.sketchpad.experiment.var.color_backend
        else:
            self.sketchpad.experiment.var.color_backend = color_backend
        im = t._to_pil()
        qim = QtGui.QImage(
            im.tobytes('raw', 'BGRA'),
            im.size[0], im.size[1],
            QtGui.QImage.Format_ARGB32
        )
        pixmap = QtGui.QPixmap.fromImage(qim)
        i = self.addPixmap(pixmap)
        c = i.boundingRect().center()
        dx = c.x()
        dy = c.y()
        if center:
            x -= dx
            y -= dy
        i.setPos(x, y)
        return i

    def line(self, sx, sy, ex, ey, color=None, penwidth=None, add=True):
        """Mimics canvas api. See openexp._canvas.canvas."""
        i = QtWidgets.QGraphicsLineItem(self._x(sx), self._y(sy), self._x(ex),
                                        self._y(ey))
        i.setPen(self._pen(color, penwidth))
        if add:
            self.addItem(i)
        return i

    def arrow(self, sx, sy, ex, ey, body_length=0.8, body_width=.5,
              head_width=30, color=None, fill=False, penwidth=None, add=True):
        """Mimics canvas api. See openexp._canvas.canvas."""
        from openexp._canvas._arrow.arrow import Arrow

        shape = Arrow._shape(self._x(sx), self._y(sy), self._x(ex),
                             self._y(ey), body_length=self._p(body_length),
                             body_width=self._p(body_width),
                             head_width=self._w(head_width))
        color = self._color(color)
        if fill:
            pen = self._pen(color, 1)
            brush = self._brush(color)
        else:
            pen = self._pen(color, penwidth)
            brush = QtGui.QBrush()
        polygon = QtGui.QPolygonF()
        for point in shape:
            polygon << QtCore.QPointF(point[0], point[1])
        i = QtWidgets.QGraphicsPolygonItem(polygon)
        i.setPen(pen)
        i.setBrush(brush)
        if add:
            self.addItem(i)
        return i

    def rect(self, x, y, w, h, fill=False, color=None, penwidth=None):
        """Mimics canvas api. See openexp._canvas.canvas."""
        color = self._color(color)
        if self._fill(fill):
            pen = self._pen(color, 1)
            brush = self._brush(color)
        else:
            pen = self._pen(color, penwidth)
            brush = QtGui.QBrush()
        i = self.addRect(self._x(x), self._y(y), self._w(w), self._h(h),
                         pen=pen, brush=brush)
        return i

    def ellipse(self, x, y, w, h, fill=False, color=None, penwidth=None,
                add=True):
        """Mimics canvas api. See openexp._canvas.canvas."""
        color = self._color(color)
        if fill:
            pen = self._pen(color, 1)
            brush = self._brush(color)
        else:
            pen = self._pen(color, penwidth)
            brush = QtGui.QBrush()
        i = QtWidgets.QGraphicsEllipseItem(self._x(x), self._y(y), self._w(w),
                                           self._h(h))
        i.setPen(pen)
        i.setBrush(brush)
        if add:
            self.addItem(i)
        return i

    def circle(self, x, y, r, fill=False, color=None, penwidth=None):
        """Mimics canvas api. See openexp._canvas.canvas."""
        r = self._r(r)
        x = self._x(x) - r
        y = self._y(y) - r
        i = self.ellipse(x, y, 2*r, 2*r, fill=fill, color=color,
                         penwidth=penwidth)
        return i

    def image(self, image, center=True, x=None, y=None, scale=None,
              rotation=None):
        """Mimics canvas api. See openexp._canvas.canvas."""
        if not isinstance(image, QtGui.QPixmap):
            image = self._pixmap(image)
        i = self.addPixmap(image)
        s = self._scale(scale)
        x = self._x(x)
        y = self._x(y)
        r = self._rotation(rotation)
        c = i.boundingRect().center()
        dx = c.x()*s
        dy = c.y()*s
        i.setTransform(
            QtGui.QTransform()
            .translate(dx, dy)
            .rotate(r)
            .translate(-dx, -dy)
            .scale(s, s)
        )
        if center:
            x -= dx
            y -= dy
        i.setPos(x, y)
        return i

    def gabor(self, x, y, *arglist, **kwdict):
        """Mimics canvas api. See openexp._canvas.canvas."""
        from openexp.canvas import gabor_file
        try:
            image = gabor_file(*arglist, **kwdict)
        except:
            self.notify(
                _(u'Some properties of a Gabor patch are unknown or variably defined, using fallback image'))
            image = self.sketchpad.theme.qpixmap(u'dialog-question')
        return self.image(image, x=x, y=y, scale=1)

    def noise_patch(self, x, y, *arglist, **kwdict):
        """Mimics canvas api. See openexp._canvas.canvas."""
        from openexp.canvas import noise_file
        try:
            image = noise_file(*arglist, **kwdict)
        except:
            self.notify(
                _(u'Some properties of a noise patch are unknown or variably defined, using fallback image'))
            image = self.sketchpad.theme.qpixmap(u'dialog-question')
        return self.image(image, x=x, y=y, scale=1)

    def fixdot(self, x=None, y=None, color=None, style=u'default'):
        """Mimics canvas api. See openexp._canvas.canvas."""
        color = self._color(color)
        x = self._x(x)
        y = self._y(y)
        group = QtWidgets.QGraphicsItemGroup()
        h = 2
        known = True
        if style == u'default':
            style = u'open-medium'
        if u'large' in style:
            s = 16
        elif u'medium' in style:
            s = 8
        elif u'small' in style:
            s = 4
        else:
            known = False
            s = 8
        if u'open' in style:
            i = self.ellipse(x-s, y-s, 2*s, 2*s, fill=True, color=color,
                             add=False)
            group.addToGroup(i)
            i = self.ellipse(x-h, y-h, 2*h, 2*h, fill=True,
                             color=self.background_color, add=False)
            group.addToGroup(i)
        elif u'filled' in style:
            i = self.ellipse(x-s, y-s, 2*s, 2*s, fill=True, color=color,
                             add=True)
            group.addToGroup(i)
        elif u'cross' in style:
            i = self.line(x, y-s, x, y+s, color=color, add=False, penwidth=1)
            group.addToGroup(i)
            i = self.line(x-s, y, x+s, y, color=color, add=False, penwidth=1)
            group.addToGroup(i)
        else:
            i = self.ellipse(x-s, y-s, 2*s, 2*s, fill=True, color=color,
                             add=True)
            group.addToGroup(i)
            known = False
        self.addItem(group)
        if not known:
            self.notify(_(u'Fixdot style "%s" is unknown or variably defined')
                        % style)
        return group


# Alias for backwards compatibility
sketchpad_canvas = SketchpadCanvas



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.dialogs.noise_settings import NoiseSettings
from libqtopensesame.sketchpad_elements._base_element import BaseElement
from libopensesame.sketchpad_elements import Noise as NoiseRuntime


class Noise(BaseElement, NoiseRuntime):

    r"""A noise element.
    See base_element for docstrings and function
    descriptions.
    """
    def show_edit_dialog(self):
        r"""The show-edit dialog for the noise shows the settings dialog."""
        d = NoiseSettings(self.sketchpad._edit_widget)
        d.set_properties(self.properties)
        properties = d.get_properties()
        if properties is None:
            return
        self.properties.update(properties)
        self.sketchpad.draw()

    @classmethod
    def mouse_press(cls, sketchpad, pos):

        d = NoiseSettings(sketchpad._edit_widget)
        properties = d.get_properties()
        if properties is None:
            return
        properties.update({
            u'x':			pos[0],
            u'y':			pos[1],
            u'show_if': 	sketchpad.current_show_if()
        })
        return noise(sketchpad, properties=properties)


# Alias for backwards compatibility
noise = Noise



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import sys
import time
from libqtopensesame.runners import BaseRunner
from qtpy import QtWidgets
from libopensesame.exceptions import UserKilled, ExperimentProcessDied
from libopensesame.oslogging import oslogger

JOIN_TIMEOUT = 3  # Seconds to wait for the process to end cleanly


class MultiprocessRunner(BaseRunner):
    """Runs an experiment in another process using multiprocessing."""
    
    supports_kill = True

    def execute(self):
        import platform
        import multiprocessing
        from libqtopensesame.misc import process, _

        self._workspace_globals = {}
        self.channel = multiprocessing.Queue()
        try:
            self.exp_process = process.ExperimentProcess(
                self.experiment,
                self.channel)
        except Exception as e:
            return e
        self.console.set_workspace_globals({'process': self.exp_process})
        # Start process!
        self.exp_process.start()
        # Wait for experiment to finish.
        # Listen for incoming messages in the meantime.
        self._finished = False
        while self.exp_process.is_alive() or not self.channel.empty():
            # We need to process the GUI. To make the GUI feel more responsive
            # during pauses, we refresh the GUI more often when paused.
            QtWidgets.QApplication.processEvents()
            if self.paused:
                for i in range(25):
                    time.sleep(.01)
                    QtWidgets.QApplication.processEvents()
            # Wait for messages. Will throw Exception if no message is received
            # before timeout.
            try:
                msg = self.channel.get(True, 0.05)
            except Exception:
                continue
            if self._process_msg(msg):
                continue
            if isinstance(msg, Exception):
                # After an exception has been captured, text and dict message
                # can still come in and need to be processed
                while True:
                    try:
                        post_msg = self.channel.get(True, .5)
                    except Exception:
                        break
                    self._process_msg(post_msg)
                # Once the process is not streaming messages anymore, close it
                self.exp_process.join(JOIN_TIMEOUT)
                try:
                    self.exp_process.close()
                except AttributeError:
                    # Process.close() was introduced only in Python 3.7
                    pass
                except ValueError:
                    # This happens when the join times out. Then we forcibly
                    # terminate the process.
                    self.exp_process.terminate()
                    oslogger.warning(
                        'experiment process was forcibly terminated')
                return msg
            # Anything that is not a string, not an Exception, and not None is
            # unexpected
            return ValueError(
                f'Illegal message type received from child process: {msg} '
                f'({type(msg)})')
        self.exp_process.join(JOIN_TIMEOUT)
        try:
            self.exp_process.close()
        except AttributeError:
            # Process.close() was introduced only in Python 3.7
            pass
        except ValueError:
            # This happens when the join times out. Then we forcibly
            # terminate the process.
            self.exp_process.terminate()
            oslogger.warning('experiment process was forcibly terminated')
        if self._finished:
            return
        if self.exp_process.killed:
            return UserKilled('The experiment process was killed.')
        return ExperimentProcessDied(
            'The experiment process seems to have died as the result of an '
            'unknown problem. This should not happen!')
        
    def _process_msg(self, msg):
        if isinstance(msg, dict):
            self._process_dict_msg(msg)
            return True
        if isinstance(msg, str):
            self._process_str_msg(msg)
            return True
        return False
        
    def _process_dict_msg(self, msg):
        # The workspace globals are sent as a dict. A special __pause__ key
        # indicates whether the experiment should be paused or resumed.
        self._workspace_globals = msg
        if '__kill__' in msg:
            self.exp_process.kill()
        if '__heartbeat__' in msg:
            self.main_window.extension_manager.fire(
                'set_workspace_globals',
                global_dict=msg
            )
            self.main_window.extension_manager.fire('heartbeat')
        elif '__pause__' in msg:
            if msg['__pause__']:
                self.pause()
            else:
                self.resume()
        elif '__finished__' in msg:
            self._finished = True
            
    def _process_str_msg(self, msg):
        sys.stdout.write(safe_decode(msg, errors='ignore'))

    def kill(self):
        self.exp_process.kill()

    def workspace_globals(self):
        return self._workspace_globals

    @staticmethod
    def has_heartbeat():
        return True



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import re
import types
from openexp.canvas_elements import ElementFactory
from libopensesame import python_workspace_api as api
from libopensesame.widgets.widget_factory import WidgetFactory
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import BaseExtension
from pyqode.core import icons
from pyqode.core.widgets import SplittableCodeEditTabWidget
from libqtopensesame.pyqode_extras.widgets import (
    PythonCodeEdit,
    FallbackCodeEdit
)
from pyqode.core.api import ColorScheme
from pyqode.core.api.panel import Panel
from pyqode.core.panels import LineNumberPanel, FoldingPanel, MarginPanel
from pyqode.core.modes import (
    PygmentsSH,
    RightMarginMode,
    CodeCompletionMode,
    AutoCompleteMode,
    BreakpointMode
)
from pyqode.python.modes import (
    sh,
    PyAutoCompleteMode,
    CalltipsMode
)
from pyqode.core.managers import backend
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import QPlainTextEdit
from qtpy.QtCore import QProcess


class PyqodeManager(BaseExtension):

    r"""Keeps track of all PyQode editor instances, which need to be explicitly
    closed. And also modifies their behavior somewhat.
    """
    def _is_opensesame_api_object(self, name, obj):

        if name.startswith(u'safe_'):
            return False
        if isinstance(obj, types.FunctionType):
            return True
        if isinstance(obj, type) and issubclass(
                obj, (WidgetFactory, ElementFactory)):
            return True
        return False

    def _register_opensesame_builtins(self):

        builtins = [
            u'exp',
            u'var',
            u'pool',
            u'items',
            u'clock',
            u'log',
            u'responses',
            u'data_files',
            u'AbortCoroutines'
        ] + [
            name
            for name, obj in api.__dict__.items()
            if self._is_opensesame_api_object(name, obj)
        ]
        sh.PythonSH.PROG = re.compile(
            sh.make_python_patterns(additional_builtins=builtins),
            re.S
        )

    def settings_widget(self):

        from .pyqode_preferences import PyQodePreferences
        return PyQodePreferences(self.main_window)

    def event_startup(self):

        backend.comm = oslogger.debug
        if self.main_window.mode == u'default':
            self._register_opensesame_builtins()
        icons.USE_QTAWESOME = True
        icons.QTA_OPTIONS = {
            'color': self.main_window.palette().text().color().name(),
            'color_disabled': self.main_window.palette().mid().color().name(),
            'color_info': '#66d9ef',
            'color_warning': '#e6db74',
            'color_error': '#f92672'
        }
        SplittableCodeEditTabWidget.editors = {}
        SplittableCodeEditTabWidget.fallback_editor = FallbackCodeEdit
        SplittableCodeEditTabWidget.register_code_edit(PythonCodeEdit)
        SplittableCodeEditTabWidget.register_code_edit(FallbackCodeEdit)
        # Syntax is broken for TextCodeEdit, so we don't use it for now
        # SplittableCodeEditTabWidget.register_code_edit(TextCodeEdit)
        self._editors = []
        self._breakpoints = {}
        self._auto_backend_restart = True

    def event_close(self):

        self._auto_backend_restart = False
        oslogger.debug(u'closing ({})'.format(len(self._editors)))
        while self._editors:
            self._editors.pop().close()
        self._auto_backend_restart = True

    def event_prepare_delete_item(self, name):

        item = self.item_store[name]
        if item.container_widget is None:
            return
        if item.item_type == u'inline_script':
            self.event_unregister_editor(item._pyqode_run_editor)
            self.event_unregister_editor(item._pyqode_prepare_editor)
        if hasattr(self, u'auto_editor'):
            for editor in self.auto_editor.values():
                self.event_unregister_editor(editor)
        self.event_unregister_editor(item._script_widget)

    def _suspend(self):

        self._auto_backend_restart = False
        for editor in self._editors:
            editor.backend.suspend()
        self._auto_backend_restart = True

    def _resume(self):

        for editor in self._editors:
            editor.backend.resume()
            self._register_backend_process(editor)

    def event_run_experiment(self, fullscreen):

        self._suspend()

    def event_end_experiment(self, ret_val):

        self._resume()

    def event_run_experiment_canceled(self):

        self._resume()

    def event_prepare_purge_unused_items(self):

        for name in self.item_store.unused():
            self.event_prepare_delete_item(name)

    def event_open_experiment(self, path):

        self.event_close()

    def event_prepare_regenerate(self):

        self.event_close()

    def event_register_editor(self, editor, mime_type=u'text/plain'):

        if editor in self._editors:
            oslogger.debug('editor already registered')
            return
        editor._auto_reset_stylesheet = False
        self._editors.append(editor)
        if 'PythonSH' in editor.modes.keys():
            pass
        elif 'TextSH' in editor.modes.keys():
            editor.modes.get('TextSH').color_scheme = safe_str(
                cfg.pyqode_color_scheme
            )
        elif 'PygmentsSH' in editor.modes.keys():
            editor.modes.get('PygmentsSH').color_scheme = safe_str(
                cfg.pyqode_color_scheme
            )
        else:
            sh = PygmentsSH(
                editor.document(),
                color_scheme=ColorScheme('monokai')
            )
            sh.set_mime_type(mime_type)
            editor.modes.append(sh)
        if cfg.pyqode_enable_breakpoints:
            editor.modes.append(BreakpointMode(breakpoints=self._breakpoints))
        editor.font_size = cfg.pyqode_font_size
        editor.font_name = cfg.pyqode_font_name
        editor.setLineWrapMode(
            QPlainTextEdit.WidgetWidth
            if cfg.pyqode_line_wrap
            else QPlainTextEdit.NoWrap
        )
        editor.action_duplicate_line.setShortcut(
            cfg.pyqode_shortcut_duplicate_line
        )
        editor.action_swap_line_up.setShortcut(
            cfg.pyqode_shortcut_swap_line_up
        )
        editor.action_swap_line_down.setShortcut(
            cfg.pyqode_shortcut_swap_line_down
        )
        editor._auto_reset_stylesheet = True
        editor._reset_stylesheet()
        editor.setTabStopWidth(
            QFontMetrics(
                editor.font()
            ).averageCharWidth() * cfg.pyqode_tab_length
        )
        # This function is defined here so to keep the editor in the scope.
        # This is necessary in case the backend crashes and it needs to be
        # restarted.

        def on_error():
            if editor._closed or not self._auto_backend_restart:
                return
            oslogger.warning(
                u'restarting backend: {}'.format(editor.__class__.__name__)
            )
            if editor.backend._share_id in editor.backend.SHARE_COUNT:
                del editor.backend.SHARE_COUNT[editor.backend._share_id]
            editor.backend._process = None
            editor.backend.resume()
            editor.backend._process.errorOccurred.connect(on_error)
            editor.backend._process.finished.connect(on_finished)
            self._register_backend_process(editor)

        def on_finished(exit_status):
            if exit_status == QProcess.CrashExit:
                oslogger.warning(
                    'backend crashed (exit status: {})'.format(exit_status))
                on_error()
        try:
            editor.backend._process.errorOccurred.connect(on_error)
            editor.backend._process.finished.connect(on_finished)
            self._register_backend_process(editor)
        except RuntimeError:
            on_error()
        oslogger.debug(u'registering {}'.format(editor))

    def event_unregister_editor(self, editor):

        if editor is None:
            return
        self._auto_backend_restart = False
        editor.close()
        if editor in self._editors:
            self._editors.remove(editor)
        oslogger.debug(u'unregistering ({})'.format(len(self._editors)))
        self._auto_backend_restart = True

    def event_setting_changed(self, setting, value):

        fnc = '_setting_{}'.format(setting)
        if hasattr(self, fnc):
            getattr(self, fnc)(value)

    def provide_pyqode_breakpoints(self):

        return self._breakpoints

    def event_pyqode_suspend_auto_backend_restart(self):

        self._auto_backend_restart = False

    def event_pyqode_resume_auto_backend_restart(self):

        self._auto_backend_restart = True

    def event_pyqode_set_breakpoints(self, breakpoints):

        self._breakpoints = breakpoints

    def event_pyqode_clear_breakpoints(self):

        self.extension_manager.fire(
            'notify',
            message=_(u'Cleared {} breakpoint(s) in {} file(s)'.format(
                sum(len(lines) for lines in self._breakpoints.values()),
                len(self._breakpoints))
            )
        )
        self._breakpoints.clear()
        for editor in self._editors:
            if 'BreakpointMode' not in editor.modes.keys():
                continue
            editor.modes.get('BreakpointMode').clear_messages()
            editor.repaint()

    def _register_backend_process(self, editor):

        self.extension_manager.fire(
            'register_subprocess',
            pid=editor.backend._process.processId(),
            description='BackendProcess:{}'.format(
                editor.__class__.__name__
            )
        )

    def _setting_pyqode_show_line_numbers(self, show_line_numbers):

        self._toggle_panel(
            'pyqode_show_line_numbers',
            show_line_numbers,
            LineNumberPanel,
            Panel.Position.LEFT
        )

    def _setting_pyqode_line_wrap(self, line_wrap):

        cfg.pyqode_line_wrap = line_wrap
        for editor in self._editors:
            editor.setLineWrapMode(
                QPlainTextEdit.WidgetWidth
                if cfg.pyqode_line_wrap
                else QPlainTextEdit.NoWrap
            )

    def _setting_pyqode_code_completion(self, code_completion):

        editors = self._editors
        self._editors = [
            e for e in editors
            if isinstance(e, FallbackCodeEdit) and
            e.language not in cfg.pyqode_code_completion_excluded_mimetypes
        ]
        for mode in (
                CodeCompletionMode,
                AutoCompleteMode
        ):
            self._toggle_mode('pyqode_code_completion', code_completion, mode)
        self._editors = [
            e for e in editors
            if isinstance(e, PythonCodeEdit) and
            e.language not in cfg.pyqode_code_completion_excluded_mimetypes
        ]
        for mode in (
                CodeCompletionMode,
                PyAutoCompleteMode,
                CalltipsMode
        ):
            self._toggle_mode('pyqode_code_completion', code_completion, mode)
        self._editors = editors

    def _setting_pyqode_fixed_width(self, fixed_width):

        self._toggle_panel(
            'pyqode_fixed_width',
            fixed_width,
            MarginPanel,
            Panel.Position.LEFT,
            nchar=cfg.pyqode_fixed_width_nchar
        )

    def _setting_pyqode_show_whitespaces(self, show_whitespaces):

        cfg.pyqode_show_whitespaces = show_whitespaces
        for editor in self._editors:
            editor.show_whitespaces = cfg.pyqode_show_whitespaces

    def _setting_pyqode_right_margin(self, show_right_margin):

        self._toggle_mode(
            'pyqode_right_margin',
            show_right_margin,
            RightMarginMode
        )

    def _setting_pyqode_code_folding(self, code_folding):

        self._toggle_panel(
            'pyqode_code_folding',
            code_folding,
            FoldingPanel,
            Panel.Position.LEFT
        )

    def event_pyqode_select_indentation_mode(self):

        def _select_indentation_mode(mode):

            cfg.pyqode_indentation = mode
            for editor in self._editors:
                if 'AutodetectIndentationMode' not in editor.modes.keys():
                    continue
                editor.modes.get(
                    'AutodetectIndentationMode')._autodetect_indentation()

        haystack = [
            (_('autodetect'), 'autodetect', _select_indentation_mode),
            (_('spaces'), 'spaces', _select_indentation_mode),
            (_('tabs'), 'tabs', _select_indentation_mode)
        ]
        self.extension_manager.fire(
            u'quick_select',
            haystack=haystack,
            placeholder_text=_(u'Select indentation mode …')
        )

    def _toggle_mode(self, config, enabled, cls, **kwargs):

        cfg[config] = enabled
        for editor in self._editors:
            if enabled:
                if cls.__name__ in editor.modes.keys():
                    continue
                editor.modes.append(cls(**kwargs))
            else:
                if cls.__name__ not in editor.modes.keys():
                    continue
                editor.modes.remove(cls.__name__)
            editor.viewport().repaint()

    def _toggle_panel(self, config, enabled, cls, pos, **kwargs):

        cfg[config] = enabled
        for editor in self._editors:
            if enabled:
                if cls.__name__ in editor.panels._panels[pos]:
                    continue
                panel = cls(**kwargs)
                editor.panels.append(panel, position=pos)
                editor.panels.refresh()
                panel.show()
            else:
                if cls.__name__ not in editor.panels._panels[pos]:
                    continue
                editor.panels.remove(cls.__name__)
                editor.panels.refresh()
